import hail as hl
from hail import Table
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def plt_hist(expression: hl.Expression, bins: int = 50, range: list = None, threshold: float = None,
             title: str = None, x_label: str = None, y_label: str = None, log: bool = False, figsize: tuple = (12, 8)):
    exprs = expression.collect()
    df = pd.DataFrame({'col': exprs})

    title = f'{title}' if title else ''
    fig = plt.figure(figsize=figsize)
    if log is True:
        df = df[df['col'] != 0]
        plt.hist(np.log10(df.col), edgecolor='black', density=False, bins=bins, color='tab:blue')
    else:
        plt.hist(df['col'], edgecolor='black', density=False, bins=bins, color='tab:blue')
    if threshold:
        plt.axvline(x=threshold, color='red', linestyle='--')
    if range is not None:
        plt.xlim(xmin=range[0], xmax=range[1])
    plt.title(title, fontsize=20)
    plt.ylabel(y_label if y_label else 'Frequency', fontsize=15)
    plt.xlabel(x_label if x_label else '', fontsize=15)
    plt.close()

    return fig


def fstat_plot(df_female, df_male, f_stat_x: float = 0.4, f_stat_y: float = 0.8, figsize: tuple = (12, 8)):
    fig, axs = plt.subplots(2, figsize=figsize)
    axs[0].hist(df_female['filters'], bins=30, histtype='bar', alpha=0.8, fill=True, color='tab:blue', edgecolor="k")
    axs[0].axvline(x=f_stat_y, color='red', linestyle='--')
    axs[0].set_title('Female Fstat', fontsize=20)
    axs[0].set_xlabel(xlabel='Fstat', fontsize=15)
    axs[0].set_ylabel(ylabel='Frequency', fontsize=15)
    axs[0].tick_params(axis='both', which='major', labelsize=12)
    axs[1].hist(df_male['filters'], bins=40, histtype='bar', alpha=0.8, fill=True, color='tab:blue', edgecolor="k")
    axs[1].axvline(x=f_stat_x, color='red', linestyle='--')
    axs[1].set_title('Male Fstat', fontsize=20)
    axs[1].set_xlabel(xlabel='Fstat', fontsize=15)
    axs[1].set_ylabel(ylabel='Frequency', fontsize=15)
    axs[1].tick_params(axis='both', which='major', labelsize=12)
    fig.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.close()

    return fig


def qqplot(pvals, title: str = None, figsize: tuple = (10, 10)):
    source = pvals._indices.source
    if isinstance(source, Table):
        ht = source.select(p_value=pvals)
    else:
        ht = source.select_rows(p_value=pvals).rows()

    ht = ht.key_by().select('p_value').key_by('p_value').persist()
    lambda_gc = hl.lambda_gc(ht['p_value'])
    n = ht.count()
    ht = ht.annotate(
        observed_p=-hl.log10(ht['p_value']),
        expected_p=-hl.log10((hl.scan.count() + 1) / n),
        p_val=ht['p_value']
    ).persist()

    p_val_pd = ht.to_pandas()
    p_val_pd['observed_p'].values[p_val_pd['observed_p'] > 10] = 10
    mini = min(p_val_pd['expected_p'].max(), p_val_pd['observed_p'].max())
    maxi = max(p_val_pd['expected_p'].max(), p_val_pd['observed_p'].max())

    title = f'{title}' if title else 'QQ Plot'

    fig = plt.figure(figsize=figsize)
    plt.scatter(p_val_pd['expected_p'], p_val_pd['observed_p'], c='black', s=0.5)
    plt.plot((0, mini), (0, mini), 'red')
    plt.xlim([0, maxi + 0.5])
    plt.ylim([0, maxi + 0.5])
    plt.title(title, fontsize=20)
    plt.ylabel('Observed -log10(' + r'$p$' + ')', fontsize=15)
    plt.xlabel('Expected -log10(' + r'$p$' + ')', fontsize=15)
    plt.close()

    return fig, round(lambda_gc, 3)


def manhattan_plot(pvals, significance_threshold: float = -np.log10(5E-08), title: str = None,
                   figsize: tuple = (17, 11), annotate_sig: bool = False):
    source = pvals._indices.source

    if isinstance(source, Table):
        ht = source.select(p_value=pvals)
    else:
        ht = source.select_rows(p_value=pvals).rows()

    data = ht.to_pandas()
    data = data.drop('alleles', 1)  # remove the 'allele' column
    data['locus'] = data['locus'].astype(str)
    data[['CHROM', 'POS']] = data.locus.str.split(":", expand=True)
    data.columns = ['locus', 'p', 'chromosome', 'position'] # rename columns
    data['position'] = data['position'].astype(int)
    data['chromosome'].replace({"X": 23, "Y": 24, "MT": 25}, inplace=True)
    data['chromosome'] = data['chromosome'].astype(int)
    data.dropna(subset=['p'], inplace=True)  # drop NAs as log10(val) won't work

    data['-log10(p_value)'] = -np.log10(data['p'])  # compute log10(pvals)
    data['-log10(p_value)'].values[data['-log10(p_value)'] > 199] = 199
    data['chromosome'] = data['chromosome'].astype('category')
    data['ind'] = range(len(data))
    data_grouped = data.groupby('chromosome')

    title = f'{title}' if title else 'Manhattan Plot'

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot()

    x_labels = []
    x_labels_pos = []

    colors = ['tab:orange', 'tab:blue']

    for num, (name, group) in enumerate(data_grouped):
        group.plot(kind='scatter', x='ind', y='-log10(p_value)', marker='o', color=colors[int(name) % len(colors)],
                   ax=ax, s=10)
        x_labels.append(name)
        x_labels_pos.append((group['ind'].iloc[-1] - (group['ind'].iloc[-1] - group['ind'].iloc[0]) / 2))

    ax.set_xticks(x_labels_pos)
    ax.set_xticklabels(x_labels)
    # ax.set_xlim([0, len(data)])
    ax.margins(0.05)
    # ax.set_ylim([0, data['-log10(p_value)'].max() + 1])
    ax.set_xlabel('Chromosome', fontsize=15)
    ax.set_ylabel('-log10(p_value)', fontsize=15)
    plt.title(title, fontsize=20)
    plt.axhline(y=significance_threshold, color='red', linestyle='--', linewidth=2)
    plt.xticks(fontsize=10, rotation=90)
    plt.yticks(fontsize=10)
    if annotate_sig is True:
        for index, row in data.iterrows():
            if row['-log10(p_value)'] >= significance_threshold:
                ax.annotate('{}:{}'.format(row['chromosome'], row['position']),
                            xy=(index, row['-log10(p_value)'] + 0.1),
                            bbox=dict(boxstyle="round", fc="0.8"))
    plt.close()

    return fig



__author__ = 'Dan King'

import hail as hl


def variant_qc_aggregator(mt) -> hl.MatrixTable:
    """:func:`.variant_qc` as an aggregator."""
    bound_exprs = {}
    gq_dp_exprs = {}

    def has_field_of_type(name, dtype):
        return name in mt.entry and mt[name].dtype == dtype

    if has_field_of_type('DP', hl.tint32):
        gq_dp_exprs['dp_stats'] = hl.agg.stats(mt.DP).select('mean', 'stdev', 'min', 'max')
    if has_field_of_type('GQ', hl.tint32):
        gq_dp_exprs['gq_stats'] = hl.agg.stats(mt.GQ).select('mean', 'stdev', 'min', 'max')
    if not has_field_of_type('GT', hl.tcall):
        raise ValueError("'variant_qc': expect an entry field 'GT' of type 'call'")
    bound_exprs['n_called'] = hl.agg.count_where(hl.is_defined(mt['GT']))
    bound_exprs['n_not_called'] = hl.agg.count_where(hl.is_missing(mt['GT']))
    n_cols = hl.agg.count()
    bound_exprs['n_filtered'] = hl.int64(n_cols) - hl.agg.count()
    bound_exprs['call_stats'] = hl.agg.call_stats(mt.GT, mt.alleles)
    return hl.rbind(hl.struct(**bound_exprs),
                    lambda e1: hl.rbind(
                        hl.case().when(hl.len(mt.alleles) == 2,
                                       hl.hardy_weinberg_test(e1.call_stats.homozygote_count[0],
                                                              e1.call_stats.AC[1] - 2
                                                              * e1.call_stats.homozygote_count[1],
                                                              e1.call_stats.homozygote_count[1])
                                       ).or_missing(),
                        lambda hwe: hl.struct(**{
                            **gq_dp_exprs,
                            **e1.call_stats,
                            'call_rate': hl.float(e1.n_called) / (e1.n_called + e1.n_not_called + e1.n_filtered),
                            'n_called': e1.n_called,
                            'n_not_called': e1.n_not_called,
                            'n_filtered': e1.n_filtered,
                            'n_het': e1.n_called - hl.sum(e1.call_stats.homozygote_count),
                            'n_non_ref': e1.n_called - e1.call_stats.homozygote_count[0],
                            'het_freq_hwe': hwe.het_freq_hwe,
                            'p_value_hwe': hwe.p_value})))


def agg_call_rate(mt: hl.MatrixTable):
    # DOES NOT HANDLE filter_entries CORRECTLY!
    n_called = hl.agg.count_where(hl.is_defined(mt['GT']))

    return hl.agg.filter(
        ~(mt.exclude_row | mt.exclude_col),
        n_called / hl.agg.count())


def impute_sex_aggregator(call,
                          aaf,
                          aaf_threshold=0.0,
                          include_par=False,
                          female_threshold=0.4,
                          male_threshold=0.8) -> hl.Table:
    """:func:`.impute_sex` as an aggregator."""
    mt = call._indices.source
    rg = mt.locus.dtype.reference_genome
    x_contigs = hl.literal(
        hl.eval(
            hl.map(lambda x_contig: hl.parse_locus_interval(x_contig, rg), rg.x_contigs)))
    inbreeding = hl.agg.inbreeding(call, aaf)
    is_female = hl.if_else(inbreeding.f_stat < female_threshold,
                           True,
                           hl.if_else(inbreeding.f_stat > male_threshold,
                                      False,
                                      hl.is_missing('tbool')))
    expression = hl.struct(is_female=is_female, **inbreeding)
    if not include_par:
        interval_type = hl.tarray(hl.tinterval(hl.tlocus(rg)))
        par_intervals = hl.literal(rg.par, interval_type)
        expression = hl.agg.filter(
            ~par_intervals.any(lambda par_interval: par_interval.contains(mt.locus)),
            expression)
    expression = hl.agg.filter((aaf > aaf_threshold) & (aaf < (1 - aaf_threshold)), expression)
    expression = hl.agg.filter(
        x_contigs.any(lambda contig: contig.contains(mt.locus)),
        expression)

    return expression


def allele_types(mt):
    from hail.expr.functions import _num_allele_type, _allele_types
    allele_types = _allele_types[:]
    allele_types.extend(['Transition', 'Transversion'])
    allele_enum = {i: v for i, v in enumerate(allele_types)}
    allele_ints = {v: k for k, v in allele_enum.items()}

    def allele_type(ref, alt):
        return hl.bind(lambda at: hl.if_else(at == allele_ints['SNP'],
                                             hl.if_else(hl.is_transition(ref, alt),
                                                        allele_ints['Transition'],
                                                        allele_ints['Transversion']),
                                             at),
                       _num_allele_type(ref, alt))

    return mt.alleles[1:].map(lambda alt: allele_type(mt.alleles[0], alt))



__author__ = 'Lindo Nkambule'

import ntpath


def get_vcf_filebase(file: str = None):

    vcf_name = ntpath.basename(file)
    if vcf_name.endswith('.gz'):
        file_no_ext = vcf_name[:-7]
    elif vcf_name.endswith('.bgz'):
        file_no_ext = vcf_name[:-8]
    else:
        file_no_ext = vcf_name[:-4]

    return file_no_ext



import hail as hl
from gwaspy.utils.read_file import read_infile


def liftover_to_grch38(
        input_type: str = None,
        dirname: str = None,
        basename: str = None):

    lifted_over = f'{dirname}{basename}.liftover.grch38.mt'
    print('\nLifting over to GRCh38')
    mt = read_infile(input_type=input_type, dirname=dirname, basename=basename)

    rg37 = hl.get_reference('GRCh37')
    rg38 = hl.get_reference('GRCh38')
    rg37.add_liftover('gs://hail-common/references/grch37_to_grch38.over.chain.gz', rg38)

    mt = mt.annotate_rows(new_locus=hl.liftover(mt.locus, 'GRCh38', include_strand=True), old_locus=mt.locus)
    mt = mt.filter_rows(hl.is_defined(mt.new_locus) & ~mt.new_locus.is_negative_strand)

    mt = mt.key_rows_by(locus=mt.new_locus.result, alleles=mt.alleles)

    print(f'\nWriting out data lifted-over to GRCh38 to: {lifted_over}')
    mt.write(lifted_over)

    return hl.read_matrix_table(lifted_over)



import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from typing import Tuple
import random


def assign_population_pcs(
        pop_pc_pd: pd.DataFrame,
        num_pcs: int,
        known_col: str = 'SuperPop',
        fit: RandomForestClassifier = None,
        seed: int = 42,
        prop_train: float = 0.8,
        n_estimators: int = 100,
        min_prob: float = 0.9,
        output_col: str = 'pop',
        missing_label: str = 'oth'
) -> Tuple[pd.DataFrame, RandomForestClassifier]:
    """
    This function uses a random forest model to assign population labels based on the results of PCA.
    Default values for model and assignment parameters are those used in gnomAD.
    :param Table pop_pc_pd: Pandas dataframe containing population PCs as well as a column with population labels
    :param str known_col: Column storing the known population labels
    :param RandomForestClassifier fit: fit from a previously trained random forest model (i.e., the output from a previous RandomForestClassifier() call)
    :param int num_pcs: number of population PCs on which to train the model
    :param int seed: Random seed
    :param float prop_train: Proportion of known data used for training
    :param int n_estimators: Number of trees to use in the RF model
    :param float min_prob: Minimum probability of belonging to a given population for the population to be set (otherwise set to `None`)
    :param str output_col: Output column storing the assigned population
    :param str missing_label: Label for samples for which the assignment probability is smaller than `min_prob`
    :return: Dataframe containing sample IDs and imputed population labels, trained random forest model
    :rtype: DataFrame, RandomForestClassifier
    """

    print(f'{num_pcs} PCs to be used in population assignment')
    # Expand PC column
    pc_cols = ['PC{}'.format(i + 1) for i in range(num_pcs)]
    train_data = pop_pc_pd.loc[~pop_pc_pd[known_col].isnull()]

    N = len(train_data)

    # Split training data into subsamples for fitting and evaluating
    if not fit:
        random.seed(seed)
        train_subsample_ridx = random.sample(list(range(0, N)), int(N * prop_train))
        train_fit = train_data.iloc[train_subsample_ridx]
        fit_samples = [x for x in train_fit['s']]
        evaluate_fit = train_data.loc[~train_data['s'].isin(fit_samples)]

        # Train RF
        training_set_known_labels = train_fit[known_col].values
        training_set_pcs = train_fit[pc_cols].values
        evaluation_set_pcs = evaluate_fit[pc_cols].values

        pop_clf = RandomForestClassifier(n_estimators=n_estimators, random_state=seed)
        pop_clf.fit(training_set_pcs, training_set_known_labels)
        print('Random forest feature importances are as follows: {}'.format(pop_clf.feature_importances_))

        # Evaluate RF
        predictions = pop_clf.predict(evaluation_set_pcs)
        error_rate = 1 - sum(evaluate_fit[known_col] == predictions) / float(len(predictions))
        print('Estimated error rate for RF model is {}'.format(error_rate))
    else:
        pop_clf = fit

    # Classify data
    print('Classifying data')
    pop_pc_pd[output_col] = pop_clf.predict(pop_pc_pd[pc_cols].values)
    probs = pop_clf.predict_proba(pop_pc_pd[pc_cols].values)
    probs = pd.DataFrame(probs, columns=[f'prob_{p}' for p in pop_clf.classes_])

    pop_pc_pd = pd.concat([pop_pc_pd.reset_index(drop=True), probs.reset_index(drop=True)], axis=1)

    probs['max'] = probs.max(axis=1)
    pop_pc_pd.loc[probs['max'] < min_prob, output_col] = missing_label

    return pop_pc_pd, pop_clf


__author__ = 'Lindo Nkambule'

import hailtop.batch as hb
import hail as hl
import pandas as pd
from gwaspy.phasing.get_filebase import get_vcf_filebase
from gwaspy.utils.get_file_size import bytes_to_gb
from typing import Union


def eagle_phasing(b: hb.batch.Batch,
                  vcf_file: str = None,
                  ref_vcf_file: hb.ResourceFile = None,
                  ref_size: Union[float, int] = None,
                  reference: str = 'GRCh38',
                  cpu: int = 8,
                  img: str = 'docker.io/lindonkambule/gwaspy:v1',
                  threads: int = 7,
                  out_dir: str = None):

    vcf_size = bytes_to_gb(vcf_file)
    mem = 'highmem' if vcf_size > 1 else 'standard'
    disk_size = round(5.0 + 3.0 * vcf_size) + round(5.0 + 3.0 * ref_size) if ref_vcf_file else round(5.0 + 3.0 * vcf_size)

    vcf_filename_no_ext = get_vcf_filebase(vcf_file)
    output_file_name = f'{vcf_filename_no_ext}.phased.withref.eagle' if ref_vcf_file else \
        f'{vcf_filename_no_ext}.phased.eagle'
    vcf = b.read_input(vcf_file)

    map_file = '/opt/genetic_map_hg38_withX.txt.gz' if reference == 'GRCh38' else '/opt/genetic_map_hg19_withX.txt.gz'

    phase = b.new_job(name=f'phase-{vcf_file}')
    phase.cpu(cpu)
    phase.memory(mem)
    phase.storage(f'{disk_size}Gi')
    phase.image(img)

    if ref_vcf_file:
        phase.command('echo INDEXING REFERENCE FILE')
        phase.command(f'bcftools index {ref_vcf_file}')
        cmd = f'''
        eagle \
            --geneticMapFile {map_file} \
            --numThreads {threads} \
            --outPrefix {output_file_name} \
            --vcfOutFormat b \
            --vcfRef {ref_vcf_file} \
            --vcfTarget {vcf}
        '''

    else:
        cmd = f'''
        eagle \
            --geneticMapFile {map_file} \
            --numThreads {threads} \
            --outPrefix {output_file_name} \
            --vcfOutFormat b \
            --vcf {vcf}
        '''

    phase.command('echo INDEXING INPUT FILE')
    phase.command(f'bcftools index {vcf}')
    phase.command('echo RUNNING PHASING')
    phase.command(cmd)
    # index the output
    phase.command(f'bcftools index {output_file_name}.bcf')

    phase.command(f'mv {output_file_name}.bcf {phase.ofile}')
    phase.command(f'mv {output_file_name}.bcf.csi {phase.csi}')
    b.write_output(phase.ofile, f'{out_dir}/{output_file_name}.bcf')
    b.write_output(phase.csi, f'{out_dir}/{output_file_name}.bcf.csi')


def shapeit_phasing(b: hb.batch.Batch,
                    vcf_file: str = None,
                    ref_vcf_file: hb.ResourceFile = None,
                    ref_size: Union[float, int] = None,
                    reference: str = 'GRCh38',
                    region: Union[str, int] = None,
                    map_chromosome: str = None,
                    cpu: int = 8,
                    img: str = 'docker.io/lindonkambule/gwaspy:v1',
                    threads: int = 7,
                    out_dir: str = None):

    vcf_size = bytes_to_gb(vcf_file)
    mem = 'highmem' if vcf_size > 1 else 'standard'
    disk_size = round(5.0 + 3.0 * vcf_size) + round(5.0 + 3.0 * ref_size) if ref_vcf_file else round(5.0 + 3.0 * vcf_size)

    vcf_filename_no_ext = get_vcf_filebase(vcf_file)
    output_file_name = f'{vcf_filename_no_ext}.phased.withref.shapeit.bcf'if ref_vcf_file else \
        f'{vcf_filename_no_ext}.phased.shapeit.bcf'
    vcf = b.read_input(vcf_file)

    chrom = map_chromosome if reference == 'GRCh38' else f'chr{map_chromosome}'

    if reference == 'GRCh37':
        if chrom == 'chr23':
            chrom = 'chrX'

    map_file = f'/shapeit4/maps/b38/{chrom}.b38.gmap.gz' if reference == 'GRCh38'\
        else f'/shapeit4/maps/b37/{chrom}.b37.gmap.gz'

    phase = b.new_job(name=f'phase-{vcf_file}')
    phase.cpu(cpu)
    phase.memory(mem)
    phase.storage(f'{disk_size}Gi')
    phase.image(img)

    if ref_vcf_file:
        phase.command('echo INDEXING REFERENCE FILE')
        phase.command(f'bcftools index {ref_vcf_file}')
        cmd = f'''
        shapeit4.2 \
            --input {vcf} \
            --map {map_file} \
            --region {region} \
            --reference {ref_vcf_file} \
            --output {output_file_name} \
            --thread {threads}
        '''

    else:
        cmd = f'''
        shapeit4.2 \
            --input {vcf} \
            --map {map_file} \
            --region {region} \
            --output {output_file_name} \
            --thread {threads}
        '''

    phase.command('echo INDEXING INPUT FILE')
    phase.command(f'bcftools index {vcf}')
    phase.command('echo RUNNING PHASING')
    phase.command(cmd)
    # index the output
    phase.command(f'bcftools index {output_file_name}')

    phase.command(f'mv {output_file_name} {phase.ofile}')
    phase.command(f'mv {output_file_name}.csi {phase.csi}')
    b.write_output(phase.ofile, f'{out_dir}/{output_file_name}')
    b.write_output(phase.csi, f'{out_dir}/{output_file_name}.csi')


def run_phase(backend: Union[hb.ServiceBackend, hb.LocalBackend] = None,
              input_vcf: str = None,
              vcf_ref_path: str = None,
              software: str = 'shapeit',
              reference: str = 'GRCh38',
              cpu: int = 8,
              threads: int = 7,
              out_dir: str = None):

    # error handling
    global ref_path, ref_chr, chrom
    if software.lower() not in ['eagle', 'shapeit']:
        raise SystemExit(f'Incorrect software {software} selected. Options are [eagle, shapeit]')

    if reference not in ['GRCh37', 'GRCh38']:
        raise SystemExit(f'Incorrect reference genome build {reference} selected. Options are [GRCh37, GRCh38]')

    vcf_filebase = get_vcf_filebase(input_vcf)
    phasing_b = hb.Batch(backend=backend,
                         name=f'phasing-{vcf_filebase}-{software}')

    if vcf_ref_path:
        if vcf_ref_path == 'hgdp_1kg':
            print(f'\n2. PHASING {input_vcf} WITH HGDP + 1000 GENOMES REFERENCE PANEL\n')
            ref_path = 'gs://gcp-public-data--gnomad/resources/hgdp_1kg/phased_haplotypes/hgdp.tgp.gwaspy.merged.chrCNUMBER.merged.bcf'
        else:
            print(f'\n2. PHASING {input_vcf} WITH USER-DEFINED REFERENCE PANEL\n')
            ref_path = vcf_ref_path
    else:
        print(f'\n2. PHASING {input_vcf} WITHOUT A REFERENCE PANEL\n')

    # get the regions so we can map each file to its specific region
    regions = pd.read_csv(f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/regions.lines', sep='\t', names=['reg', 'ind'])
    regions_dict = pd.Series(regions.reg.values, index=regions.ind).to_dict()

    scatter_vcfs_paths = hl.utils.hadoop_ls(f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/scatter_vcfs/*.bcf')

    vcfs = []
    for i in scatter_vcfs_paths:
        vcfs.append(i['path'])

    phased_vcf_out_dir = f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/phased_scatter'

    for i in range(1, 24):
        if reference == 'GRCh38':
            if i == 23:
                chrom = 'chrX'
                ref_chr = 'X'
            else:
                chrom = f'chr{i}'
                ref_chr = i

        else:
            chrom = str(i)
            ref_chr = i

        ref_chrom_path = ref_path.replace('CNUMBER', str(ref_chr)) if vcf_ref_path else None
        ref_vcf_chrom_file = phasing_b.read_input(ref_chrom_path) if vcf_ref_path else None
        ref_vcf_chrom_file_size = bytes_to_gb(ref_chrom_path) if vcf_ref_path else None

        for file in vcfs:
            # get specific region for file using regions.line file
            vcf_basename = get_vcf_filebase(file)
            file_index = int(vcf_basename.split('.')[-1])
            file_region = regions_dict[file_index]
            map_chrom = file_region.split(':')[0]

            if map_chrom == chrom:
                if software == 'eagle':
                    # check if file exists to avoid re-doing things
                    out_phased_filename = f'{vcf_basename}.phased.withref.eagle.bcf' if ref_vcf_chrom_file else \
                        f'{vcf_basename}.phased.eagle.bcf'

                    if hl.hadoop_exists(f'{phased_vcf_out_dir}/{out_phased_filename}'):
                        continue
                    else:
                        eagle_phasing(b=phasing_b, vcf_file=file, ref_vcf_file=ref_vcf_chrom_file,
                                      reference=reference, ref_size=ref_vcf_chrom_file_size, cpu=cpu, threads=threads,
                                      out_dir=phased_vcf_out_dir)

                else:
                    out_phased_filename = f'{vcf_basename}.phased.withref.shapeit.bcf' if ref_vcf_chrom_file else \
                        f'{vcf_basename}.phased.shapeit.bcf'

                    # check if file exists to avoid re-doing things
                    if hl.hadoop_exists(f'{phased_vcf_out_dir}/{out_phased_filename}'):
                        continue
                    else:
                        shapeit_phasing(b=phasing_b, vcf_file=file, ref_vcf_file=ref_vcf_chrom_file,
                                        ref_size=ref_vcf_chrom_file_size, reference=reference, region=file_region,
                                        map_chromosome=map_chrom, cpu=cpu, threads=threads, out_dir=phased_vcf_out_dir)

    phasing_b.run()





__author__ = 'Lindo Nkambule'

import hailtop.batch as hb
import hail as hl
import pandas as pd
from typing import Union
from gwaspy.phasing.get_filebase import get_vcf_filebase
from gwaspy.utils.get_file_size import bytes_to_gb


def imputation(b: hb.batch.Batch,
               vcf: str = None,
               vcf_filename_no_ext: str = None,
               ref: hb.ResourceGroup = None,
               ref_size: Union[int, float] = None,
               region: str = None,
               chromosome: str = None,
               cpu: int = 8,
               memory: str = 'highmem',
               img: str = 'docker.io/lindonkambule/gwaspy:v1',
               threads: int = 7,
               out_dir: str = None):

    # in_vcf = b.read_input(vcf)
    in_vcf = b.read_input_group(**{'bcf': vcf,
                                'bcf.csi': f'{vcf}.csi'})
    vcf_size = bytes_to_gb(vcf)

    output_file_name = vcf_filename_no_ext + '.imputed.bcf'
    file_dir = vcf_filename_no_ext.split('.')[0]

    disk_size = ref_size + (vcf_size * 4)

    map_file = f'/shapeit4/maps/b38/{chromosome}.b38.gmap.gz'

    impute = b.new_job(name=output_file_name)
    impute.cpu(cpu)
    impute.memory(memory)
    impute.storage(f'{disk_size}Gi')
    impute.image(img)

    cmd = f'''
        impute5_1.1.5_static \
            --h {ref.bcf} \
            --m {map_file} \
            --g {in_vcf.bcf} \
            --r {region} \
            --out-gp-field \
            --o {output_file_name} \
            --threads {threads}
    '''

    impute.command(cmd)
    # index file to use when merging
    impute.command(f'bcftools index {output_file_name}')

    impute.command(f'mv {output_file_name} {impute.ofile}')
    impute.command(f'mv {output_file_name}.csi {impute.ind}')
    b.write_output(impute.ofile, f'{out_dir}/GWASpy/Imputation/{file_dir}/imputed_chunks/{output_file_name}')
    b.write_output(impute.ind, f'{out_dir}/GWASpy/Imputation/{file_dir}/imputed_chunks/{output_file_name}.csi')


def run_impute(backend: Union[hb.ServiceBackend, hb.LocalBackend] = None,
               input_vcfs: str = None,
               phasing_software: str = None,
               memory: str = 'highmem',
               cpu: int = 8,
               threads: int = 7,
               out_dir: str = None):

    print(f'RUNNING IMPUTATION ON FILES PHASED WITH {phasing_software.upper()}')
    impute_b = hb.Batch(backend=backend, name=f'impute-phased-chunks')

    vcf_paths = pd.read_csv(input_vcfs, sep='\t', header=None)

    # get the regions so we can map each file to its specific region
    regions = pd.read_csv(f'{out_dir}/GWASpy/Phasing/regions.lines', sep='\t', names=['reg', 'ind'])
    regions_dict = pd.Series(regions.reg.values, index=regions.ind).to_dict()

    for index, row in vcf_paths.iterrows():
        vcf = row[0]
        vcf_filebase = get_vcf_filebase(vcf)

        if phasing_software == 'shapeit':
            phased_vcfs_chunks = hl.utils.hadoop_ls(f'{out_dir}/GWASpy/Phasing/{vcf_filebase}/phased_scatter/*.shapeit.bcf')
        else:
            phased_vcfs_chunks = hl.utils.hadoop_ls(f'{out_dir}/GWASpy/Phasing/{vcf_filebase}/phased_scatter/*.eagle.bcf')

        for i in range(1, 24):
            if i == 23:
                chrom = 'chrX'
            else:
                chrom = f'chr{i}'

            ref_bcf = f'gs://gcp-public-data--gnomad/resources/hgdp_1kg/phased_haplotypes/hgdp.tgp.gwaspy.merged.{chrom}.merged.bcf'
            ref_ind = f'gs://gcp-public-data--gnomad/resources/hgdp_1kg/phased_haplotypes/hgdp.tgp.gwaspy.merged.{chrom}.merged.bcf.csi'
            ref_size = bytes_to_gb(ref_bcf)
            ref = impute_b.read_input_group(**{'bcf': ref_bcf,
                                               'bcf.csi': ref_ind})

            for file in phased_vcfs_chunks:
                f = file['path']
                vcf_basename = get_vcf_filebase(f)
                file_index = int(vcf_basename.split('.')[-3])
                file_region = regions_dict[file_index]
                map_chrom = file_region.split(':')[0]

                imp_out_filename = f'{vcf_basename}.imputed.bcf'
                file_dir = vcf_basename.split('.')[0]
                output_filepath_name = f'{out_dir}/GWASpy/Imputation/{file_dir}/imputed_chunks/{imp_out_filename}'

                if map_chrom == chrom:
                    # check if imputed file already exists
                    if hl.hadoop_exists(output_filepath_name):
                        continue
                    else:
                        imputation(b=impute_b, vcf=f, vcf_filename_no_ext=vcf_basename, ref=ref, ref_size=ref_size,
                                   region=file_region, chromosome=chrom, cpu=cpu, memory=memory,
                                   threads=threads, out_dir=out_dir)

    impute_b.run()



__author__ = 'Lindo Nkambule'

import hail as hl


def pca_filter_mt(
        in_mt: hl.MatrixTable,
        maf: float = 0.05,
        hwe: float = 1e-3,
        call_rate: float = 0.98,
        ld_cor: float = 0.2,
        ld_window: int = 250000):

    print("\nInitial number of SNPs before filtering: {}".format(in_mt.count_rows()))
    mt = hl.variant_qc(in_mt)
    print(f'\nFiltering out variants with MAF < {maf}')
    mt_filt = mt.annotate_rows(maf=hl.min(mt.variant_qc.AF))
    mt_filt = mt_filt.filter_rows(mt_filt.maf > maf)

    print(f'\nFiltering out variants with HWE < {hwe:1e}')
    mt_filt = mt_filt.filter_rows(mt_filt.variant_qc.p_value_hwe > hwe)

    print(f'\nFiltering out variants with Call Rate < {call_rate}')
    mt_filt = mt_filt.filter_rows(mt_filt.variant_qc.call_rate >= call_rate)

    # no strand ambiguity
    print('\nFiltering out strand ambigous variants')
    mt_filt = mt_filt.filter_rows(~hl.is_strand_ambiguous(mt_filt.alleles[0], mt_filt.alleles[1]))

    # MHC chr6:25-35Mb
    # chr8.inversion chr8:7-13Mb
    print('\nFiltering out variants in MHC [chr6:25M-35M] and chromosome 8 inversions [chr8:7M-13M]')
    intervals = ['chr6:25M-35M', 'chr8:7M-13M']
    mt_filt = hl.filter_intervals(mt_filt, [hl.parse_locus_interval(x, reference_genome='GRCh38') for x in intervals],
                                  keep=False)

    # This step is expensive (on local machine)
    print(f'\nLD pruning using correlation threshold of {ld_cor} and window size of {ld_window}')
    mt_ld_prune = hl.ld_prune(mt_filt.GT, r2=ld_cor, bp_window_size=ld_window)
    mt_ld_pruned = mt_filt.filter_rows(hl.is_defined(mt_ld_prune[mt_filt.row_key]))
    print("\nNumber of SNPs after filtering: {}".format(mt_ld_pruned.count_rows()))

    return mt_ld_pruned


def relatedness_check(
        in_mt: hl.MatrixTable = None,
        method: str = 'pc_relate',
        outdir: str = None,
        kin_estimate: float = 0.98):

    global mt, samples_to_remove

    in_mt = hl.variant_qc(in_mt)
    in_mt = hl.sample_qc(in_mt)

    # _localize=False means don't put this in Python, keep it as a Hail expr
    call_rate_dict = in_mt.aggregate_cols(hl.dict(hl.agg.collect((in_mt.s, in_mt.sample_qc.call_rate))), _localize=False)

    if method == 'pc_relate':
        print("\nUsing PC-Relate for relatedness checks")
        relatedness_ht = hl.pc_relate(in_mt.GT, 0.01, k=10, min_kinship=0.1, statistics='kin')
        samples_to_remove_ht = relatedness_ht.filter(relatedness_ht.kin > kin_estimate)

        # get call rates for both samples so we remove the one with lower call rate between the two
        samples_to_remove = samples_to_remove_ht.annotate(cr_s1=call_rate_dict[samples_to_remove_ht.i.s],
                                                          cr_s2=call_rate_dict[samples_to_remove_ht.j.s])

        samples_list = samples_to_remove.annotate(sample_to_remove=hl.cond(
            samples_to_remove.cr_s1 <= samples_to_remove.cr_s2, samples_to_remove.i, samples_to_remove.j))

    elif method == 'ibd':
        print("\nUsing PLINK-style identity by descent for relatedness checks")
        in_mt = in_mt.annotate_rows(maf=hl.min(in_mt.variant_qc.AF))
        relatedness_ht = hl.identity_by_descent(in_mt, maf=in_mt['maf'])  # this returns a Hail Table with the sample pairs
        samples_to_remove_ht = relatedness_ht.filter(relatedness_ht.ibd.PI_HAT > kin_estimate)

        # get call rates for both samples so we remove the one with lower call rate between the two
        samples_to_remove = samples_to_remove_ht.annotate(cr_s1=call_rate_dict[samples_to_remove_ht.i],
                                                          cr_s2=call_rate_dict[samples_to_remove_ht.j])

        samples_list = samples_to_remove.annotate(sample_to_remove=hl.cond(
            samples_to_remove.cr_s1 <= samples_to_remove.cr_s2, samples_to_remove.i, samples_to_remove.j))

    else:
        print("\nUsing KING for relatedness checks")
        if kin_estimate > 0.5:
            raise Exception("\nThe maximum kinship coefficient is for KING 0.5")
        relatedness_mt = hl.king(in_mt.GT)
        filtered_relatedness_mt = relatedness_mt.filter_entries((relatedness_mt.s_1 != relatedness_mt.s) &
                                                                (relatedness_mt.phi >= kin_estimate), keep=True)
        samples_to_remove_ht = filtered_relatedness_mt.entries()
        samples_to_remove = samples_to_remove_ht.annotate(cr_s1=call_rate_dict[samples_to_remove_ht.s_1],
                                                          cr_s2=call_rate_dict[samples_to_remove_ht.s])

        samples_list = samples_to_remove.annotate(sample_to_remove=hl.cond(
            samples_to_remove.cr_s1 <= samples_to_remove.cr_s2, samples_to_remove.s_1, samples_to_remove.s))

    samples = samples_list.sample_to_remove.collect()

    if len(samples) > 0:
        in_mt = in_mt.filter_cols(hl.literal(samples).contains(in_mt['s']), keep=False)
        print("\nNumber of samples that fail relatedness checks: {}".format(len(samples)))
        with open(outdir + 'relatedness_removed_samples.tsv', 'w') as f:
            for sample in samples:
                f.write(sample + "\n")

    else:
        print("\nNo samples failed the relatedness check")

    return in_mt



__author__ = 'Lindo Nkambule'

import hailtop.batch as hb
import hail as hl
import pandas as pd
from gwaspy.utils.get_file_size import bytes_to_gb
from gwaspy.phasing.get_filebase import get_vcf_filebase
from typing import List
from typing import Union


def concat_vcfs(b: hb.batch.Batch,
                vcf_basename: str = None,
                vcfs_to_merge: List = None,
                output_type: str = 'vcf',
                chrom: str = None,
                cpu: int = 16,
                memory: str = 'standard',
                docker_img: str = 'docker.io/lindonkambule/gwaspy:v1',
                out_dir: str = None):

    global index_cmd

    out_type = 'b' if output_type == 'bcf' else 'z'
    vcfs_sizes_sum = 0
    merge_vcf_i = ''

    out_filename = f'{vcf_basename}.{chrom}.merged.bcf' if output_type == 'bcf' else \
        f'{vcf_basename}.{chrom}.merged.vcf.gz'
    out_index_name = f'{vcf_basename}.{chrom}.merged.bcf.csi' if output_type == 'bcf' else \
        f'{vcf_basename}.{chrom}.merged.vcf.gz.csi'

    for line in vcfs_to_merge:
        vcfs_sizes_sum += 2 + bytes_to_gb(line)

    disk_size = int(round(10 + (2 * vcfs_sizes_sum)))
    threads = cpu - 1

    concat = b.new_job(name=f'concat-{vcf_basename}')
    concat.memory(memory)
    concat.storage(f'{disk_size}Gi')
    concat.image(docker_img)
    concat.cpu(cpu)

    for line in vcfs_to_merge:
        input_vcf = b.read_input_group(vcf=line,
                                       ind=f'{line}.csi')
        merge_vcf_i += f'{input_vcf.vcf} '

    cmd = f'''
        bcftools concat \
            --no-version \
            --output-type {out_type} \
            --output {out_filename} \
            --threads {threads} \
            {merge_vcf_i}
    '''

    concat.command(cmd)
    # index the merged output
    concat.command(f'bcftools index --force {out_filename}')

    concat.command(f'mv {out_filename} {concat.ofile}')
    concat.command(f'mv {out_index_name} {concat.idx}')
    b.write_output(concat.ofile, f'{out_dir}/GWASpy/{vcf_basename}/Imputation/imputed_merged/{out_filename}')
    b.write_output(concat.idx, f'{out_dir}/GWASpy/{vcf_basename}/Imputation/imputed_merged/{out_index_name}')


def run_concat(backend: Union[hb.ServiceBackend, hb.LocalBackend] = None,
               input_vcf: str = None,
               output_type: str = 'vcf',
               cpu: int = 16,
               memory: str = 'standard',
               out_dir: str = None):

    print(f'\n2. CONCAT {input_vcf}\n')
    vcf_filebase = get_vcf_filebase(input_vcf)
    concat_b = hb.Batch(backend=backend, name=f'concat-imputed-chunks-{vcf_filebase}')

    # get the regions so we can map each file to its specific region
    regions = pd.read_csv(f'{out_dir}/GWASpy/{vcf_filebase}/Imputation/imputation.regions', sep='\t', names=['reg', 'ind'])
    regions_dict = pd.Series(regions.reg.values, index=regions.ind).to_dict()

    imputed_vcfs_chunks = hl.utils.hadoop_ls(f'{out_dir}/GWASpy/{vcf_filebase}/Imputation/imputed_chunks/*.bcf')

    for i in range(1, 24):
        if i == 23:
            chrom = 'chrX'
        else:
            chrom = f'chr{i}'

        chrom_phased_files_to_concat = []

        for file in imputed_vcfs_chunks:
            f = file['path']
            vcf_basename = get_vcf_filebase(f)
            file_index = int(vcf_basename.split('.')[-4])
            file_region = regions_dict[file_index]
            map_chrom = file_region.split(':')[0]
            if map_chrom == chrom:
                chrom_phased_files_to_concat.append(f)

        # naturally sort the list of files to merge
        from gwaspy.utils.natural_sort import natural_keys
        chrom_phased_files_to_concat.sort(key=natural_keys)

        concat_vcfs(b=concat_b, vcfs_to_merge=chrom_phased_files_to_concat, vcf_basename=vcf_filebase,
                    output_type=output_type, chrom=chrom, cpu=cpu, memory=memory, out_dir=out_dir)

    concat_b.run()



__author__ = 'Michael Wilson & Lindo Nkambule'

import hailtop.batch as hb
import argparse


def haplotype_phasing(input_vcf: str = None,
                      vcf_ref: str = None,
                      local: bool = False,
                      billing_project: str = None,
                      software: str = 'shapeit',
                      reference: str = 'GRCh38',
                      max_win_size_cm: float = 10.0,
                      overlap_size_cm: float = 2.0,
                      scatter_memory: int = 26,
                      cpu: int = 8,
                      threads: int = 7,
                      stages: str = 'scatter,phase,concat',
                      output_type: str = 'bcf',
                      out_dir: str = None):
    # Error handling
    if not out_dir:
        raise SystemExit('Output directory not specified. Specify using --out_dir if running from the command line or'
                         'out_dir argument if running inside a Python script')

    steps_list = stages.split(',')
    steps = [x.lower() for x in steps_list]
    unknown_steps = [i for i in steps if i not in ['scatter', 'phase', 'concat']]

    if len(unknown_steps) > 0:
        raise SystemExit(f'Incorrect process(es) {unknown_steps} selected. Options are [scatter, phase, concat]')

    if output_type.lower() not in ['bcf', 'vcf']:
        raise SystemExit(f'Incorrect output type {output_type} selected. Options are [bcf, vcf]')

    if local:
        backend = hb.LocalBackend()
    else:
        backend = hb.ServiceBackend(billing_project=billing_project,
                                    remote_tmpdir=f'{out_dir}/tmp/')

    # Scatter VCF/BCF file(s)
    if 'scatter' in steps:
        from gwaspy.phasing.scatter_vcf import run_scatter
        run_scatter(backend=backend, input_vcf=input_vcf, reference=reference, max_win_size_cm=max_win_size_cm,
                    overlap_size_cm=overlap_size_cm, scatter_memory=scatter_memory, out_dir=out_dir)

    # Phase scatterd chunks
    if 'phase' in steps:
        from gwaspy.phasing.phase_vcf import run_phase
        run_phase(backend=backend, input_vcf=input_vcf, vcf_ref_path=vcf_ref, software=software, reference=reference,
                  cpu=cpu, threads=threads, out_dir=out_dir)

    # Concatenate phased chunks
    if 'concat' in steps:
        from gwaspy.phasing.concat_vcfs import run_concat
        run_concat(backend=backend, input_vcf=input_vcf, output_type=output_type, reference=reference,
                   software=software, out_dir=out_dir)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input-vcf', type=str, required=True)
    parser.add_argument('--vcf-ref', type=str, default=None)
    parser.add_argument('--local', action='store_true')
    parser.add_argument('--billing-project', required=True)
    parser.add_argument('--software', type=str, default='shapeit', choices=['eagle', 'shapeit'])
    parser.add_argument('--reference', type=str, default='GRCh38', choices=['GRCh37', 'GRCh38'])
    parser.add_argument('--max-win-size-cm', type=float, default=10.0)
    parser.add_argument('--overlap-size-cm', type=float, default=2.0)
    parser.add_argument('--cpu', type=int, default=8)
    parser.add_argument('--scatter-mem', type=int, default=26)
    parser.add_argument('--threads', type=int, default=7)
    parser.add_argument('--stages', type=str, default='scatter,phase,concat')
    parser.add_argument('--out-type', type=str, default='bcf', choices=['bcf', 'vcf'])
    parser.add_argument('--out-dir', required=True)

    args = parser.parse_args()

    haplotype_phasing(input_vcf=args.input_vcf, vcf_ref=args.vcf_ref, local=args.local,
                      billing_project=args.billing_project, software=args.software, reference=args.reference,
                      max_win_size_cm=args.max_win_size_cm, overlap_size_cm=args.overlap_size_cm,
                      scatter_memory=args.scatter_mem, cpu=args.cpu, threads=args.threads, stages=args.stages,
                      output_type=args.out_type, out_dir=args.out_dir)


if __name__ == '__main__':
    main()



__author__ = 'Lindo Nkambule'

from pylatex import Document, Section, Subsection, Command, Center, Tabular, NewPage, Figure, SubFigure, TextColor
from pylatex.utils import NoEscape, bold


class MyDocument(Document):
    def __init__(self, basename):
        super().__init__()

        self.preamble.append(Command('title', 'QC Report of {}'.format(basename)))
        # self.preamble.append(Command('author', 'Anonymous author'))
        self.preamble.append(Command('date', NoEscape(r'\today')))
        self.append(NoEscape(r'\maketitle'))

    def flags_table(self, pre_qc_counts=None, pos_qc_counts=None, results=None, lambda_gc=None, sig_vars=None):
        nids_lost = (pre_qc_counts['n_samples'] - pos_qc_counts['n_samples']) / pre_qc_counts['n_samples']
        nids_lost = round(nids_lost, 4)
        nids_sex_check = results['sex_warnings'][True] / pre_qc_counts['n_samples']
        nids_sex_check = round(nids_sex_check, 4)

        if 'is_case_counts' in pre_qc_counts.keys() | pos_qc_counts.keys():
            cas_con_ratio = round(pos_qc_counts['is_case_counts']['case'] / pos_qc_counts['is_case_counts']['control'], 4)

            tbl = [['Number of SNPs Post-QC', pos_qc_counts['n_variants'], 250000, 200000, 0, 'green'],
                   ['Number of Cases Post-QC', pos_qc_counts['is_case_counts']['case'], 100, 50, 0, 'green'],
                   ['Number of Controls Post-QC', pos_qc_counts['is_case_counts']['control'], 100, 50, 0, 'green'],
                   ['Case-Control ratio Post-QC', cas_con_ratio, 0.0625, 0.0278, 0, 'green'],
                   ['Number of IDs lost ratio', nids_lost, 0.01, 0.1, 0, 'green'],
                   ['Number of IDs with no Phenotype Post-QC', pos_qc_counts['is_case_counts']['unknown'], 0, 10, 0, 'green'],
                   ['Ratio of IDs that failed sex checks', nids_sex_check, 0.005, 0.025, 0, 'green'],
                   ['Lambda GC Post-QC', lambda_gc, 1.1, 1.2, 0, 'green'],
                   ['Number of Significant GWAS hits Post-QC', sig_vars, 0, 1, 0, 'green']]

        else:
            tbl = [['Number of SNPs Post-QC', pos_qc_counts['n_variants'], 250000, 200000, 0, 'green'],
                   ['Number of IDs lost ratio', nids_lost, 0.01, 0.1, 0, 'green'],
                   ['Ratio of IDs that failed sex checks', nids_sex_check, 0.005, 0.025, 0, 'green']]

        for i in tbl:
            if (i[0] == 'Number of SNPs Post-QC') | (i[0] == 'Number of Cases Post-QC') |\
                    (i[0] == 'Number of Controls Post-QC') | (i[0] == 'Case-Control ratio Post-QC'):
                if (i[1] <= i[2]) & (i[1] >= i[3]):
                    i[4] = 1
                    i[5] = 'orange'
                elif i[1] < i[3]:
                    i[4] = 2
                    i[5] = 'red'
                else:
                    i[4] = 0
                    i[5] = 'green'

            if i[0] == 'Number of IDs lost ratio':
                if (i[1] > i[2]) & (i[1] <= i[3]):
                    i[4] = 1
                    i[5] = 'orange'
                elif i[1] > i[3]:
                    i[4] = 2
                    i[5] = 'red'
                else:
                    i[4] = 0
                    i[5] = 'green'

            if (i[0] == 'Ratio of IDs that failed sex checks') | (i[0] == 'Lambda GC Post-QC'):
                if (i[1] >= i[2]) & (i[1] <= i[3]):
                    i[4] = 1
                    i[5] = 'orange'
                elif i[1] > i[3]:
                    i[4] = 2
                    i[5] = 'red'
                else:
                    i[4] = 0
                    i[5] = 'green'

            if (i[0] == 'Number of Significant GWAS hits Post-QC') | (i[0] == 'Number of IDs with no Phenotype Post-QC'):
                if i[1] == i[2]:
                    i[4] = 0
                    i[5] = 'green'
                elif (i[1] > i[2]) & (i[1] <= i[3]):
                    i[4] = 1
                    i[5] = 'orange'
                else:
                    i[4] = 2
                    i[5] = 'red'

        with self.create(Section('Flags')):
            with self.create(Center()) as centered:
                with centered.create(Tabular('|l|c|c|c|c|c|')) as table:
                    table.add_hline()
                    table.add_row((bold('Flagname'), bold('Value'), bold('yellow-th'), bold('red-th'), bold('flag'),
                                   bold('color')))
                    table.add_hline()
                    for i in range(0, len(tbl)):
                        table.add_row(tbl[i][0], tbl[i][1], tbl[i][2], tbl[i][3], tbl[i][4],
                                      TextColor(tbl[i][5], tbl[i][5]))
                        table.add_hline()

    def general_info(self, pre_qc_conts, post_qc_conts, count_results, pre_filter, id_cr, fhet_thresh, var_cr,
                     miss_diff, hwe_con, hwe_cas, hwe_all, data_type):
        global pre_pheno_counts, post_pheno_counts, pheno_diffs
        if 'is_case_counts' in pre_qc_conts.keys() | post_qc_conts.keys():
            pre_pheno_counts = str(pre_qc_conts['is_case_counts']['case']) + ', ' + \
                               str(pre_qc_conts['is_case_counts']['control']) + ', ' + \
                               str(pre_qc_conts['is_case_counts']['unknown'])

            post_pheno_counts = str(post_qc_conts['is_case_counts']['case']) + ', ' + \
                                str(post_qc_conts['is_case_counts']['control']) + ', ' + \
                                str(post_qc_conts['is_case_counts']['unknown'])

            pheno_diffs = str(pre_qc_conts['is_case_counts']['case'] - post_qc_conts['is_case_counts']['case']) + ', ' \
                          + str(pre_qc_conts['is_case_counts']['control'] - post_qc_conts['is_case_counts']['control'])\
                          + ', ' +\
                          str(pre_qc_conts['is_case_counts']['unknown'] - post_qc_conts['is_case_counts']['unknown'])

        pre_sex_counts = str(pre_qc_conts['is_female_counts']['male']) + ', ' + \
                         str(pre_qc_conts['is_female_counts']['female']) + ', ' + \
                         str(pre_qc_conts['is_female_counts']['unknown'])

        post_sex_counts = str(post_qc_conts['is_female_counts']['male']) + ', ' + \
                          str(post_qc_conts['is_female_counts']['female']) + ', ' + \
                          str(post_qc_conts['is_female_counts']['unknown'])

        sex_diffs = str(pre_qc_conts['is_female_counts']['male']-post_qc_conts['is_female_counts']['male']) + ', '\
                    + str(pre_qc_conts['is_female_counts']['female']-post_qc_conts['is_female_counts']['female']) +\
                    ', ' + str(pre_qc_conts['is_female_counts']['unknown']-post_qc_conts['is_female_counts']['unknown'])

        with self.create(Section('General Info')):
            # with self.create(Subsection('Size of sample')):
            with self.create(Center()) as centered:
                with centered.create(Tabular('|c|c|c|c|')) as table:
                    table.add_hline()
                    table.add_row((bold('Test'), bold('pre QC'), bold('post QC'), bold('exlcusion-N')))
                    table.add_hline()
                    if 'is_case_counts' in pre_qc_conts.keys() | post_qc_conts.keys():
                        table.add_row(('Cases, Controls, Missing', pre_pheno_counts,
                                        post_pheno_counts, pheno_diffs))
                        table.add_hline()
                    table.add_row(('Males, Females, Unspec', pre_sex_counts,
                                    post_sex_counts, sex_diffs))
                    table.add_hline()
                    table.add_row(('SNPs', pre_qc_conts['n_variants'],
                                    post_qc_conts['n_variants'],
                                    pre_qc_conts['n_variants'] - post_qc_conts['n_variants']))
                    table.add_hline()

            # with self.create(Subsection('Exclusion overview')):
            with self.create(Center()) as centered:
                with centered.create(Tabular('|l|l|')) as table:
                    table.add_hline()
                    table.add_row((bold('Filter'), bold('N ')))
                    table.add_hline()
                    table.add_row(('SNPs: call rate < {} (pre - filter)'.format(pre_filter),
                                    count_results['pre_geno'][True]))
                    table.add_row(('IDs: call rate (cases/controls) < {}'.format(id_cr),
                                    count_results['mind'][True]))
                    table.add_row(('IDs: FHET outside +- {} (cases/controls)'.format(fhet_thresh),
                                    count_results['fstat'][True]))
                    table.add_row(('IDs: Sex violations -excluded- (N-tested)', count_results['sex_violations'][True]))
                    table.add_row(
                        ('IDs: Sex warnings (undefined phenotype / ambiguous genotypes)',
                            count_results['sex_warnings'][True]))
                    table.add_row(('SNPs: call rate < {}'.format(var_cr), count_results['geno'][True]))
                    if data_type != "no-pheno":
                        table.add_row(('SNPs: missing diference > {}'.format(miss_diff), count_results['cr_diff'][True]))
                    table.add_row(('SNPs: without valid association p-value (invariant)',
                                   count_results['monomorphic_var'][True]))
                    if data_type == "Case-only":
                        table.add_row(('SNPs: HWE-cases < {}'.format(hwe_cas), count_results['hwe_cas'][True]))
                        table.add_hline()
                    if data_type == "Control-only":
                        table.add_row(('SNPs: HWE-controls < {}'.format(hwe_con), count_results['hwe_con'][True]))
                        table.add_hline()
                    if data_type == "Case-Control":
                        table.add_row(('SNPs: HWE-controls < {}'.format(hwe_con), count_results['hwe_con'][True]))
                        table.add_row(('SNPs: HWE-cases < {}'.format(hwe_cas), count_results['hwe_cas'][True]))
                        table.add_hline()
                    if data_type == "no-pheno":
                        table.add_row(('SNPs: HWE < {}'.format(hwe_all), count_results['hwe_all'][True]))
                        table.add_hline()

    def manhattan_sec(self, qq_pre_path, qq_pos_path, man_pre_path, man_pos_path, table_results):
        self.append(NewPage())

        with self.create(Section('Manhattan')):
            with self.create(Subsection('Basic stats')):
                with self.create(Center()) as centered:
                    with centered.create(Tabular('|c|c|c|')) as table:
                        table.add_hline()
                        table.add_row((bold('Description'), bold('Pre-QC'), bold('Post-QC')))
                        table.add_hline()
                        table.add_row(('Number of GWAS hits', table_results[0], table_results[1]))
                        table.add_hline()
                        table.add_row(('Lambda GC', table_results[2], table_results[3]))
                        table.add_hline()
                        # calculate lambda 1000
                        table.add_row(('Lambda 1000', table_results[4], table_results[5]))
                        table.add_hline()

            self.append(NewPage())

            with self.create(Subsection('Manhattan Plot - pre-QC')):
                with self.create(Figure(position='h!')) as pre_man_images:
                    self.append(Command('centering'))
                    with self.create(
                            SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as man_pre_images:
                        man_pre_images.add_image(man_pre_path, width=NoEscape(r'1\linewidth'))
                with self.create(Figure(position='h!')) as pre_man_images:
                    self.append(Command('centering'))
                    with self.create(
                            SubFigure(position='c', width=NoEscape(r'0.85\linewidth'))) as qq_pre_images:
                        qq_pre_images.add_image(qq_pre_path, width=NoEscape(r'0.85\linewidth'))

            self.append(NewPage())

            with self.create(Subsection('Manhattan Plot - post-QC')):
                with self.create(Figure(position='h!')) as pos_man_images:
                    self.append(Command('centering'))
                    with self.create(
                            SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as man_pos_images:
                        man_pos_images.add_image(man_pos_path, width=NoEscape(r'1\linewidth'))
                with self.create(Figure(position='h!')) as pos_man_images:
                    self.append(Command('centering'))
                    with self.create(
                            SubFigure(position='c', width=NoEscape(r'0.85\linewidth'))) as qq_pos_images:
                        qq_pos_images.add_image(qq_pos_path, width=NoEscape(r'0.85\linewidth'))

    def individual_char(self, id_con_pre_path, id_cas_pre_path, id_all_path, fstat_fig_path, data_type):
        self.append(NewPage())

        with self.create(Section('Per Individual Characteristics Analysis')):
            with self.create(Subsection('Missing Rates - pre-QC')):
                if data_type == "Case-only":
                    with self.create(Figure(position='h!')) as pre_idcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as idcr_cas_pre_images:
                            idcr_cas_pre_images.add_image(id_cas_pre_path, width=NoEscape(r'1\linewidth'))
                    self.append(NewPage())

                if data_type == "Control-only":
                    with self.create(Figure(position='h!')) as pre_idcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as idcr_con_pre_images:
                            idcr_con_pre_images.add_image(id_con_pre_path, width=NoEscape(r'1\linewidth'))
                    self.append(NewPage())

                if data_type == "no-pheno":
                    with self.create(Figure(position='h!')) as pre_idcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as idcr_all_pre_images:
                            idcr_all_pre_images.add_image(id_all_path, width=NoEscape(r'1\linewidth'))
                    self.append(NewPage())

                if data_type == "Case-Control":
                    with self.create(Figure(position='h!')) as pre_idcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as idcr_con_pre_images:
                            idcr_con_pre_images.add_image(id_con_pre_path, width=NoEscape(r'1\linewidth'))
                    with self.create(Figure(position='h!')) as pre_idcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as idcr_cas_pre_images:
                            idcr_cas_pre_images.add_image(id_cas_pre_path, width=NoEscape(r'1\linewidth'))
                    self.append(NewPage())

            with self.create(Subsection('Fstat - Sex Violations')):
                with self.create(Figure(position='h!')) as fstat_image:
                    self.append(Command('centering'))
                    with self.create(
                            SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as fstat_fig:
                        fstat_fig.add_image(fstat_fig_path, width=NoEscape(r'1\linewidth'))

    def snp_char(self, var_con_pre_path, var_cas_pre_path, var_all_path, data_type):
        self.append(NewPage())

        with self.create(Section('Per SNP Characteristics Analysis')):
            with self.create(Subsection('Missing Rates - pre-QC')):
                if data_type == "Case-only":
                    with self.create(Figure(position='h!')) as pre_varcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as varcr_cas_pre_images:
                            varcr_cas_pre_images.add_image(var_cas_pre_path, width=NoEscape(r'1\linewidth'))
                    self.append(NewPage())

                if data_type == "Control-only":
                    with self.create(Figure(position='h!')) as pre_varcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as varcr_con_pre_images:
                            varcr_con_pre_images.add_image(var_con_pre_path, width=NoEscape(r'1\linewidth'))
                    self.append(NewPage())

                if data_type == "no-pheno":
                    with self.create(Figure(position='h!')) as pre_varcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as varcr_all_pre_images:
                            varcr_all_pre_images.add_image(var_all_path, width=NoEscape(r'1\linewidth'))
                    self.append(NewPage())

                if data_type == "Case-Control":
                    with self.create(Figure(position='h!')) as pre_varcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as varcr_con_pre_images:
                            varcr_con_pre_images.add_image(var_con_pre_path, width=NoEscape(r'1\linewidth'))
                    with self.create(Figure(position='h!')) as pre_varcr_images:
                        self.append(Command('centering'))
                        with self.create(
                                SubFigure(position='c', width=NoEscape(r'1\linewidth'))) as varcr_cas_pre_images:
                            varcr_cas_pre_images.add_image(var_cas_pre_path, width=NoEscape(r'1\linewidth'))
                    self.append(NewPage())




__author__ = 'Lindo Nkambule'

import hail as hl
import sys


def add_sample_annotations(mt: hl.MatrixTable, annotations: str) -> hl.MatrixTable:
    # use annotations file to annotate VCF
    ann = hl.import_table(annotations, impute=False,
                          types={'Sample': hl.tstr, 'Sex': hl.tstr, 'Pheno': hl.tstr}).key_by('Sample')
    ann_cols = dict(ann.row)

    mt = mt.annotate_cols(annotations=ann[mt.s])

    if 'is_female' not in mt.col:
        if 'Sex' in ann_cols:
            mt = mt.annotate_cols(is_female=hl.if_else(((mt.annotations.Sex == 'F') |
                                                        (mt.annotations.Sex == str(2)) |
                                                        (mt.annotations.Sex == 'True') |
                                                        (mt.annotations.Sex == 'Female')),
                                                       True, False))
        else:
            print('Sex column is missing from annotations file. Please add it and run GWASpy again')
            sys.exit(2)

    if 'is_case' not in mt.col:
        if 'Pheno' in ann_cols:
            mt = mt.annotate_cols(is_case=hl.if_else(((mt.annotations.Pheno == str(2)) |
                                                      (mt.annotations.Pheno == 'True') |
                                                      (mt.annotations.Pheno == 'Case')),
                                                     True, False))

    return mt




import hail as hl


def export_qced_file(mt: hl.MatrixTable, out_dir: str, basename: str, export_type='hail'):
    outname = basename + '_qced'

    if export_type == 'hail':
        mt.write('{}GWASpy/Preimp_QC/{}.mt'.format(out_dir, outname), overwrite=True)

    elif export_type == 'plink':
        hl.export_plink(dataset=mt, output='{}GWASpy/Preimp_QC/{}'.format(out_dir, outname), fam_id=mt.fam_id,
                        ind_id=mt.s, pat_id=mt.pat_id, mat_id=mt.mat_id, is_female=mt.is_female, pheno=mt.is_case,
                        varid=mt.rsid)

    else:
        hl.export_vcf(mt, '{}GWASpy/Preimp_QC/{}.vcf.bgz'.format(out_dir, outname))



__author__ = 'Lindo Nkambule'

import pandas as pd, numpy as np
from gwaspy.utils.get_file_size import bytes_to_gb
from gwaspy.phasing.get_filebase import get_vcf_filebase
import hailtop.batch as hb
import pathlib
import ntpath
from typing import Union


def create_windows_bed(reference: str = 'GRCh38',
                       max_win_size_cm: float = 10.0,
                       overlap_size_cm: float = 2.0,
                       vcf_filebase: str = None,
                       out_dir: str = None):
    print('creating BED file with overlapping windows to be used in splitting the input VCF')
    maps_path = 'https://storage.googleapis.com/broad-alkesgroup-public/Eagle/downloads/tables'
    if reference == 'GRCh38':
        chrom_lens = {'chr1': 248956422, 'chr2': 242193529, 'chr3': 198295559, 'chr4': 190214555, 'chr5': 181538259,
                      'chr6': 170805979, 'chr7': 159345973, 'chr8': 145138636, 'chr9': 248956422, 'chr10': 242193529,
                      'chr11': 198295559, 'chr12': 190214555, 'chr13': 181538259, 'chr14': 170805979,
                      'chr15': 159345973, 'chr16': 145138636, 'chr17': 248956422, 'chr18': 242193529,
                      'chr19': 198295559, 'chr20': 190214555, 'chr21': 181538259, 'chr22': 170805979, 'chrX': 159345973}

        df_map = pd.read_csv(f'{maps_path}/genetic_map_hg38_withX.txt.gz', delim_whitespace=True,
                             compression='gzip', header=0, names=['CHR', 'POS', 'RATE', 'CM'])
    else:
        chrom_lens = {'1': 249250621, '2': 243199373, '3': 198022430, '4': 191154276, '5': 180915260, '6': 171115067,
                      '7': 159138663, '8': 146364022, '9': 141213431, '10': 135534747, '11': 135006516, '12': 133851895,
                      '13': 115169878, '14': 107349540, '15': 102531392, '16': 90354753, '17': 81195210, '18': 78077248,
                      '19': 59128983, '20': 63025520, '21': 48129895, '22': 51304566, '23': 155270560}

        df_map = pd.read_csv(f'{maps_path}/genetic_map_hg19_withX.txt.gz', delim_whitespace=True,
                             compression='gzip', header=0, names=['CHR', 'POS', 'RATE', 'CM'])
    df_out = {}

    for chrom, df_group in df_map.groupby('CHR'):
        fai_chr = str(chrom) if str(chrom) in chrom_lens else 'chr' + str(chrom) if 'chr' + str(
            chrom) in chrom_lens else 'X' if 'X' in chrom_lens else 'chrX' if 'chrX' in chrom_lens else None
        if fai_chr:
            chr_cm_len = max(df_group['CM'])
            n_win = np.ceil((chr_cm_len - overlap_size_cm) / (max_win_size_cm - overlap_size_cm))
            win_size = (chr_cm_len - overlap_size_cm) / n_win + overlap_size_cm
            cm_begs = (win_size - overlap_size_cm) * np.arange(1, n_win)
            cm_ends = (win_size - overlap_size_cm) * np.arange(1, n_win) + overlap_size_cm
            pos_begs = np.concatenate(
                ([1], np.interp(cm_begs, df_group['CM'], df_group['POS'], period=np.inf).astype(int)))
            pos_ends = np.concatenate(
                (np.interp(cm_ends, df_group['CM'], df_group['POS'], period=np.inf).astype(int), [chrom_lens[fai_chr]]))
            df_out[fai_chr] = pd.DataFrame.from_dict({'CHR': fai_chr, 'BEG': pos_begs, 'END': pos_ends})

    df = pd.concat([df_out[fai_chr] for fai_chr in chrom_lens.keys()])

    df[['CHR', 'BEG', 'END']].to_csv(f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/refscatter.bed', sep='\t', header=False,
                                     index=False)


def vcf_scatter(b: hb.batch.Batch,
                vcf_file: str = None,
                intervals_bed: str = None,
                docker_img: str = 'docker.io/lindonkambule/gwaspy:v1',
                memory: float = 26,
                out_dir: str = None):

    global cmd

    vcf_size = bytes_to_gb(vcf_file)
    disk_size = round(10.0 + 3.0 * vcf_size)
    cpu = 2 * round(memory / 13) if memory > 6.5 else 1
    threads = cpu - 1

    vcf_filename_no_ext = get_vcf_filebase(vcf_file)

    vcf = b.read_input(vcf_file)
    bed = b.read_input(intervals_bed)

    scatter = b.new_job(name=f'scatter-{vcf_filename_no_ext}')
    scatter.cpu(cpu)
    scatter.memory(f'{memory}Gi')
    scatter.storage(f'{disk_size}Gi')
    scatter.image(docker_img)

    # work out if file is BCF or VCF
    vcf_basename = ntpath.basename(vcf_file)
    if '.vcf' in pathlib.Path(vcf_basename).suffixes:
        cmd = f'''
            set -euo pipefail
            mkdir vcfs
            awk -F"\t" '{{print $1":"$2"-"$3"\t"NR-1}}' {bed} > regions.lines
            bcftools query --list-samples "{vcf}" | tee "{vcf_filename_no_ext}.sample_id.lines" | wc -l > n_smpls.int
            bcftools view --threads {threads} --output-type u "{vcf}" | \
            bcftools annotate \
                --no-version \
                --force \
                --output-type u \
                --remove ID,QUAL,INFO,^FMT/GT \
                --threads {threads} | \
            bcftools +scatter \
                --no-version \
                --output-type b \
                --output vcfs \
                --threads {threads} \
                --scatter-file regions.lines \
                --prefix "{vcf_filename_no_ext}."
        '''
    elif '.bcf' in pathlib.Path(vcf_basename).suffixes:
        cmd = f'''
            set -euo pipefail
            mkdir vcfs
            awk -F"\t" '{{print $1":"$2"-"$3"\t"NR-1}}' {bed} > regions.lines
            bcftools query --list-samples "{vcf}" | tee "{vcf_filename_no_ext}.sample_id.lines" | wc -l > n_smpls.int
            bcftools annotate \
                --no-version \
                --force \
                --output-type u \
                --remove ID,QUAL,INFO,^FMT/GT \
                --threads {threads} \
                "{vcf}" | \
            bcftools +scatter \
                --no-version \
                --output-type b \
                --output vcfs \
                --threads {threads} \
                --scatter-file regions.lines \
                --prefix "{vcf_filename_no_ext}."
        '''

    else:
        raise SystemExit('Unsupported file format. Choices are VCF OR BCF')

    scatter.command(cmd)

    scatter.command(f'mv vcfs {scatter.vcfs}')
    scatter.command(f'mv regions.lines {scatter.regions}')
    b.write_output(scatter.vcfs, f'{out_dir}/GWASpy/{vcf_filename_no_ext}/Phasing/scatter_vcfs')
    b.write_output(scatter.regions, f'{out_dir}/GWASpy/{vcf_filename_no_ext}/Phasing/regions.lines')


def run_scatter(backend: Union[hb.ServiceBackend, hb.LocalBackend] = None,
                input_vcf: str = None,
                reference: str = 'GRCh38',
                max_win_size_cm: float = 10.0,
                overlap_size_cm: float = 2.0,
                scatter_memory: int = 26,
                out_dir: str = None):

    print(f'\n1. SCATTER {input_vcf}\n')
    vcf_filebase = get_vcf_filebase(input_vcf)

    create_windows_bed(reference=reference, max_win_size_cm=max_win_size_cm, out_dir=out_dir, vcf_filebase=vcf_filebase,
                       overlap_size_cm=overlap_size_cm)

    scatter_b = hb.Batch(backend=backend, name=f'scatter-{vcf_filebase}')

    vcf_scatter(
        b=scatter_b,
        vcf_file=input_vcf,
        intervals_bed=f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/refscatter.bed',
        memory=scatter_memory,
        out_dir=out_dir
    )

    scatter_b.run()





__author__ = 'Lindo Nkambule & Zan Koenig'

import hail as hl
from gwaspy.preimp_qc.aggregators import agg_call_rate, variant_qc_aggregator, impute_sex_aggregator
from gwaspy.preimp_qc.plots import plt_hist, fstat_plot, qqplot, manhattan_plot
import pandas as pd


class BaseFilter:
    def __init__(self):
        pass

    def filter(self, mt):
        pass

    def plot(self, mt):
        pass


class pre_geno(BaseFilter):
    def __init__(self, pre_geno_cr: float = 0.95, pre_row_filter: str = None, pre_col_filter: str = None):
        super().__init__()
        self._pre_geno_cr = pre_geno_cr
        self._row_filter = pre_row_filter
        self._col_filter = pre_col_filter

    def filter(self, mt):
        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = row_filter | col_filter

        mt = mt.annotate_rows(**{
            'pre_geno_noy': hl.struct(
                filters=hl.agg.filter(((pre_filter == False) & (mt.geno_y_excluded == True)),
                                      variant_qc_aggregator(mt).call_rate) < self._pre_geno_cr),
            'pre_geno_y': hl.struct(
                filters=hl.agg.filter(((pre_filter == False) & (mt.geno_y_only == True)),
                                      variant_qc_aggregator(mt).call_rate) < self._pre_geno_cr)})

        mt = mt.annotate_rows(**{
            'pre_geno': hl.struct(
                filters=((hl.agg.any(mt['pre_geno_noy'].filters) == True) |
                         (hl.agg.any(mt['pre_geno_y'].filters) == True))
            )})

        return mt

    def plot(self, mt):
        pass


class id_call_rate(BaseFilter):
    def __init__(self, mind: float = 0.98, pre_row_filter: str = None, pre_col_filter: str = None, data_type: str = None):
        super().__init__()
        self._mind = mind
        self._row_filter = pre_row_filter
        self._col_filter = pre_col_filter
        self._data_type = data_type

    def filter(self, mt):
        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = row_filter | col_filter

        mt = mt.annotate_cols(**{
            'mind': hl.struct(
                filters=hl.agg.filter(pre_filter == False, agg_call_rate(mt)) < self._mind)})

        return mt

    def plot(self, mt):
        global id_call_rate_plts
        mt = mt.annotate_cols(
            mind_cr_pre=hl.agg.filter(mt.pre_geno.filters == False, agg_call_rate(mt)))

        if 'is_case' in mt.col:
            if self._data_type == "Case-only":
                mt_cases: hl.MatrixTable = mt.filter_cols(mt.is_case == True)
                cas_pre = plt_hist(mt_cases.mind_cr_pre, title="Cases", threshold=self._mind, x_label='Call Rate')
                id_call_rate_plts = [cas_pre]

            if self._data_type == "Control-only":
                mt_controls: hl.MatrixTable = mt.filter_cols(mt.is_case == False)
                con_pre = plt_hist(mt_controls.mind_cr_pre, title="Controls", threshold=self._mind, x_label='Call Rate')
                id_call_rate_plts = [con_pre]

            if self._data_type == "Case-Control":
                mt_controls: hl.MatrixTable = mt.filter_cols(mt.is_case == False)
                mt_cases: hl.MatrixTable = mt.filter_cols(mt.is_case == True)

                con_pre = plt_hist(mt_controls.mind_cr_pre, title="Controls", threshold=self._mind, x_label='Call Rate')
                cas_pre = plt_hist(mt_cases.mind_cr_pre, title="Cases", threshold=self._mind, x_label='Call Rate')

                id_call_rate_plts = [con_pre, cas_pre]

        else:
            all_cas_con = plt_hist(mt.mind_cr_pre, title="Cases+Controls", threshold=self._mind, x_label='Call Rate')
            id_call_rate_plts = [all_cas_con]

        return id_call_rate_plts


class fhet_autosomes(BaseFilter):
    def __init__(self, fhet_thresh: float = 0.2, pre_row_filter: str = None,
                 pre_col_filter: str = None):
        super().__init__()
        self._fhet_th = fhet_thresh
        self._row_filter = pre_row_filter
        self._col_filter = pre_col_filter

    def filter(self, mt):
        mt = mt.annotate_rows(variant_qc=variant_qc_aggregator(mt))

        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = row_filter | col_filter

        mt = mt.annotate_cols(**{
            'fstat': hl.struct(
                filters=hl.agg.filter(pre_filter == False & mt.locus.in_autosome(),
                                      (hl.agg.inbreeding(mt.GT, hl.min(mt.variant_qc.AF)).f_stat < -self._fhet_th) |
                                      (hl.agg.inbreeding(mt.GT, hl.min(mt.variant_qc.AF)).f_stat > self._fhet_th))
            )})

        return mt

    def plot(self, mt):
        pass


class fhet_sex(BaseFilter):
    def __init__(self, fstat_x: float = 0.5, fstat_y: float = 0.5, pre_row_filter: str = None,
                 pre_col_filter: str = None, figsize: tuple = (12, 8)):
        super().__init__()
        self._fstat_x = fstat_x
        self._fstat_y = fstat_y
        self._row_filter = pre_row_filter
        self._col_filter = pre_col_filter
        self._figsize = figsize

    def filter(self, mt):
        mt = mt.annotate_rows(aaf=hl.agg.call_stats(mt.GT, mt.alleles).AF[1])
        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = row_filter | col_filter

        mt = mt.annotate_cols(**{
            'sex_violations': hl.struct(
                filters=hl.agg.filter(pre_filter == False,
                                      ((hl.agg.filter(mt.is_female == True,
                                                      impute_sex_aggregator(mt.GT, mt.aaf).f_stat)) > self._fstat_y) |
                                      ((hl.agg.filter(mt.is_female == False,
                                                      impute_sex_aggregator(mt.GT, mt.aaf).f_stat)) < self._fstat_x))
            )})

        return mt

    def plot(self, mt):
        mt = mt.annotate_cols(**{
            'fstat_sex': hl.struct(
                filters=hl.agg.filter(mt.pre_geno.filters == False, impute_sex_aggregator(mt.GT, mt.aaf).f_stat)
            )})

        mt_xy: hl.MatrixTable = mt.filter_cols(mt.is_female == True)
        mt_xx: hl.MatrixTable = mt.filter_cols(mt.is_female == False)

        exprs_xy = mt_xy.fstat_sex.collect()
        exprs_xx = mt_xx.fstat_sex.collect()

        df_xy = pd.DataFrame(exprs_xy)
        df_xx = pd.DataFrame(exprs_xx)

        fig = fstat_plot(df_female=df_xy, df_male=df_xx, f_stat_y=self._fstat_y, f_stat_x=self._fstat_x,
                         figsize=self._figsize)

        return fig


class fhet_sex_warnings(BaseFilter):
    def __init__(self, fstat_x: float = 0.8, fstat_y: float = 0.2, pre_row_filter: str = None,
                 pre_col_filter: str = None):
        super().__init__()
        self._fstat_x = fstat_x
        self._fstat_y = fstat_y
        self._row_filter = pre_row_filter
        self._col_filter = pre_col_filter

    def filter(self, mt):
        mt = mt.annotate_rows(aaf=hl.agg.call_stats(mt.GT, mt.alleles).AF[1])

        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = row_filter | col_filter

        # sex warnings are for ambiguous genotypes (F_male < 0.8, F_female > 0.2) and undefined phenotypes
        mt = mt.annotate_cols(**{
            'sex_ambiguous': hl.struct(
                filters=hl.agg.filter(pre_filter == False,
                                      ((hl.agg.filter(mt.is_female == True,
                                                      impute_sex_aggregator(mt.GT, mt.aaf).f_stat)) > self._fstat_y) |
                                      ((hl.agg.filter(mt.is_female == False,
                                                      impute_sex_aggregator(mt.GT, mt.aaf).f_stat)) < self._fstat_x))
            )})

        if 'is_case' in mt.col:
            mt = mt.annotate_cols(**{
                'sex_warnings': hl.struct(
                    filters=((hl.agg.any(mt['sex_ambiguous'].filters) == True) |
                             (hl.agg.any(hl.is_missing(mt.is_case)))
                             ))})
        else:
            mt = mt.annotate_cols(**{
                'sex_warnings': hl.struct(
                    filters=((hl.agg.any(mt['sex_ambiguous'].filters) == True)
                             ))})

        return mt

    def plot(self, mt):
        pass


class geno(BaseFilter):
    def __init__(self, geno_thresh: float = 0.98, pre_row_filter: str = None,
                 pre_col_filter: str = None, data_type: str = None):
        super().__init__()
        self._geno = geno_thresh
        self._row_filter = pre_row_filter
        self._col_filter = pre_col_filter
        self._data_type = data_type

    def filter(self, mt):
        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = row_filter | col_filter

        mt = mt.annotate_rows(**{
            'geno_noy': hl.struct(
                filters=hl.agg.filter(((pre_filter == False) & (mt.geno_y_excluded == True)),
                                      variant_qc_aggregator(mt).call_rate) < self._geno),
            'geno_y': hl.struct(
                filters=hl.agg.filter(((pre_filter == False) & (mt.geno_y_only == True)),
                                      variant_qc_aggregator(mt).call_rate) < self._geno)})

        mt = mt.annotate_rows(**{
            'geno': hl.struct(
                filters=((hl.agg.any(mt['geno_noy'].filters) == True) |
                         (hl.agg.any(mt['geno_y'].filters) == True))
            )})

        return mt

    def plot(self, mt):
        global var_call_rate_plts
        mt = mt.annotate_rows(
            var_cr_pre=hl.agg.filter(mt.pre_geno.filters == False, agg_call_rate(mt)))

        if 'is_case' in mt.col:
            if self._data_type == "Case-only":
                mt_cases: hl.MatrixTable = mt.filter_cols(mt.is_case == True)
                cas_pre = plt_hist(mt_cases.var_cr_pre, title="Cases", threshold=self._geno, x_label='Call Rate')
                var_call_rate_plts = [cas_pre]

            if self._data_type == "Control-only":
                mt_controls: hl.MatrixTable = mt.filter_cols(mt.is_case == False)
                con_pre = plt_hist(mt_controls.var_cr_pre, title="Controls", threshold=self._geno, x_label='Call Rate')
                var_call_rate_plts = [con_pre]

            if self._data_type == "Case-Control":
                mt_controls: hl.MatrixTable = mt.filter_cols(mt.is_case == False)
                mt_cases: hl.MatrixTable = mt.filter_cols(mt.is_case == True)

                con_pre = plt_hist(mt_controls.var_cr_pre, title="Controls", threshold=self._geno, x_label='Call Rate')
                cas_pre = plt_hist(mt_cases.var_cr_pre, title="Cases", threshold=self._geno, x_label='Call Rate')

                var_call_rate_plts = [con_pre, cas_pre]

        else:
            all_cas_con = plt_hist(mt.var_cr_pre, title="Cases+Controls", threshold=self._geno, x_label='Call Rate')
            var_call_rate_plts = [all_cas_con]

        return var_call_rate_plts


class call_rate_diff(BaseFilter):
    def __init__(self, pre_col_filter: str = None, pre_row_filter: str = None,
                 initial_row_filter: str = None, cr_thresh: float = 0.02):
        super().__init__()
        self._cr_thresh = cr_thresh
        self._row_filter = pre_row_filter
        self._col_filter = pre_col_filter
        self._initial_row_filter = initial_row_filter

    def filter(self, mt):
        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        mt = mt.annotate_rows(cr=hl.or_missing(row_filter == False,
                                               hl.agg.group_by(mt.is_case,
                                                               hl.agg.filter(col_filter == False,
                                                                             variant_qc_aggregator(mt).call_rate))))

        mt = mt.annotate_rows(diff=hl.abs(mt.cr[False] - mt.cr[True]))

        mt = mt.annotate_rows(**{
            'cr_diff': hl.struct(
                filters=hl.agg.any((mt.diff > self._cr_thresh) & (mt[self._initial_row_filter].filters == False)))})

        return mt

    def plot(self, mt):
        pass


class invariant(BaseFilter):
    def __init__(self, pre_col_filter: str = None):
        super().__init__()
        self._col_filter = pre_col_filter

    def filter(self, mt):
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = col_filter

        mt = mt.annotate_rows(**{
            'monomorphic_var': hl.struct(
                filters=hl.agg.filter(pre_filter == False, hl.min(variant_qc_aggregator(mt).AC)) == 0)})

        return mt

    def plot(self, mt):
        pass


class maf(BaseFilter):
    def __init__(self, maf_thresh: float = 0.01, pre_col_filter: str = None):
        super().__init__()
        self._maf_thresh = maf_thresh
        self._col_filter = pre_col_filter

    def filter(self, mt):
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = col_filter

        mt = mt.annotate_rows(**{
            'maf': hl.struct(
                filters=hl.agg.filter(pre_filter == False, hl.min(variant_qc_aggregator(mt).AF)) < self._maf_thresh)})

        return mt

    def plot(self, mt):
        pass


class hwe_con(BaseFilter):
    def __init__(self, hwe_th_co: float = 1e-06, pre_col_filter: str = None,
                 pre_row_filter: str = None):
        super().__init__()
        self._hwe_th_co = hwe_th_co
        self._col_filter = pre_col_filter
        self._row_filter = pre_row_filter

    def filter(self, mt):
        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = row_filter | col_filter

        mt = mt.annotate_rows(**{
            'hwe_con_aut': hl.struct(
                filters=((row_filter == False) &
                         (hl.agg.filter(((pre_filter == False) &
                                         (mt.is_case == False) &
                                         (mt.hwe_aut == True)),
                                        variant_qc_aggregator(mt).p_value_hwe) < self._hwe_th_co))),
            'hwe_con_sex': hl.struct(
                filters=((row_filter == False) &
                         (hl.agg.filter(((pre_filter == False) &
                                         (mt.is_case == False) &
                                         (mt.hwe_sex == True)),
                                        variant_qc_aggregator(mt).p_value_hwe) < self._hwe_th_co)))
        })

        mt = mt.annotate_rows(**{
            'hwe_con': hl.struct(
                filters=((hl.agg.any(mt['hwe_con_aut'].filters) == True) |
                         (hl.agg.any(mt['hwe_con_sex'].filters) == True))
            )})

        return mt

    def plot(self, mt):
        pass


class hwe_cas(BaseFilter):
    def __init__(self, hwe_th_ca: float = 1e-10, pre_col_filter: str = None,
                 pre_row_filter: str = None):
        super().__init__()
        self._hwe_th_ca = hwe_th_ca
        self._col_filter = pre_col_filter
        self._row_filter = pre_row_filter

    def filter(self, mt):
        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = row_filter | col_filter

        mt = mt.annotate_rows(**{
            'hwe_cas_aut': hl.struct(
                filters=((row_filter == False) &
                         (hl.agg.filter(((pre_filter == False) &
                                         (mt.is_case == True) &
                                         (mt.hwe_aut == True)),
                                        variant_qc_aggregator(mt).p_value_hwe) < self._hwe_th_ca))),
            'hwe_cas_sex': hl.struct(
                filters=((row_filter == False) &
                         (hl.agg.filter(((pre_filter == False) &
                                         (mt.is_case == True) &
                                         (mt.hwe_sex == True)),
                                        variant_qc_aggregator(mt).p_value_hwe) < self._hwe_th_ca)))
        })

        mt = mt.annotate_rows(**{
            'hwe_cas': hl.struct(
                filters=((hl.agg.any(mt['hwe_cas_aut'].filters) == True) |
                         (hl.agg.any(mt['hwe_cas_sex'].filters) == True))
            )})

        return mt

    def plot(self, mt):
        pass


class hwe_all(BaseFilter):
    def __init__(self, hwe_th_all: float = 1e-06, pre_col_filter: str = None,
                 pre_row_filter: str = None):
        super().__init__()
        self._hwe_th_all = hwe_th_all
        self._col_filter = pre_col_filter
        self._row_filter = pre_row_filter

    def filter(self, mt):
        row_filter = mt[self._row_filter].filters if self._row_filter else mt.exclude_row
        col_filter = mt[self._col_filter].filters if self._col_filter else mt.exclude_col

        pre_filter = row_filter | col_filter

        mt = mt.annotate_rows(**{
            'hwe_all_aut': hl.struct(
                filters=((row_filter == False) &
                         (hl.agg.filter(((pre_filter == False) &
                                         (mt.hwe_aut == True)),
                                        variant_qc_aggregator(mt).p_value_hwe) < self._hwe_th_all))),
            'hwe_all_sex': hl.struct(
                filters=((row_filter == False) &
                         (hl.agg.filter(((pre_filter == False) &
                                         (mt.hwe_sex == True)),
                                        variant_qc_aggregator(mt).p_value_hwe) < self._hwe_th_all)))
        })

        mt = mt.annotate_rows(**{
            'hwe_all': hl.struct(
                filters=((hl.agg.any(mt['hwe_all_aut'].filters) == True) |
                         (hl.agg.any(mt['hwe_all_sex'].filters) == True))
            )})

        return mt

    def plot(self, mt):
        pass


class manhattan(BaseFilter):
    def __init__(self, qqtitle, mantitle):
        super().__init__()
        self._qqtitle = qqtitle
        self._mantitle = mantitle

    def filter(self, mt):
        gwas = hl.linear_regression_rows(y=mt.is_case, x=mt.GT.n_alt_alleles(), covariates=[1.0])
        n_sig_variants = gwas.filter(gwas.p_value < 5E-8).count()

        return gwas, n_sig_variants

    def plot(self, ht):
        qq_plot, lambda_gc = qqplot(ht.p_value, title=self._qqtitle)
        man_plot = manhattan_plot(ht.p_value, title=self._mantitle)

        return qq_plot, lambda_gc, man_plot



__author__ = 'Lindo Nkambule'

import hailtop.batch as hb
import argparse


# ONCE YOU ADD FUNCTIONALITY FOR USING DIFFERENT REF PANELS, CHANGE n_panel* parameters and include cmd args
def genotype_imputation(input_vcf: str = None,
                        females_file: str = None,
                        n_samples: int = None,
                        n_panel_samples: int = 4099,
                        buffer_region: int = 250,
                        phasing_software: str = None,
                        local: bool = False,
                        billing_project: str = None,
                        memory: str = 'highmem',
                        cpu: int = 16,
                        stages: str = 'impute,concat',
                        output_type: str = 'bcf',
                        out_dir: str = None):
    # Error handling
    if not out_dir:
        raise SystemExit('Output directory not specified. Specify using --out_dir if running from the command line or'
                         'out_dir argument if running inside a Python script')

    steps_list = stages.split(',')
    steps = [x.lower() for x in steps_list]
    unknown_steps = [i for i in steps if i not in ['impute', 'concat']]

    if len(unknown_steps) > 0:
        raise SystemExit(f'Incorrect process(es) {unknown_steps} selected. Options are [impute, concat]')

    if output_type.lower() not in ['bcf', 'vcf']:
        raise SystemExit(f'Incorrect output type {output_type} selected. Options are [bcf, vcf]')

    if memory.lower() not in ['lowmem', 'standard', 'highmem']:
        raise SystemExit(f'Incorrect memory type {memory} selected. Options are [lowmem, standard, highmem]')

    if not n_samples:
        raise SystemExit('Number of samples in input data not detected. Specify how many samples (integer), using'
                         '--n-samples if running from the command line or'
                         'n_samples argument if running inside a Python script, are in the input data')

    if local:
        backend = hb.LocalBackend()
    else:
        backend = hb.ServiceBackend(billing_project=billing_project,
                                    remote_tmpdir=f'{out_dir}/tmp/')

    # impute genotypes
    if 'impute' in steps:
        from gwaspy.imputation.sex_aut_imp import run_impute
        run_impute(backend=backend, input_vcf=input_vcf, females_file=females_file, n_samples=n_samples,
                   n_panel_samples=n_panel_samples, phasing_software=phasing_software, memory=memory,
                   buffer_region=buffer_region, out_dir=out_dir)

    # Concatenate imputed chunks
    if 'concat' in steps:
        from gwaspy.imputation.concat_vcfs import run_concat
        run_concat(backend=backend, input_vcf=input_vcf, output_type=output_type, cpu=cpu, memory=memory,
                   out_dir=out_dir)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input-vcf', type=str, required=True)
    parser.add_argument('--samples-file', type=str, required=True)
    parser.add_argument('--local', action='store_true')
    parser.add_argument('--billing-project', required=True)
    parser.add_argument('--phasing-software', type=str, default='shapeit', choices=['eagle', 'shapeit'])
    parser.add_argument('--memory', type=str, default='highmem', choices=['lowmem', 'standard', 'highmem'])
    parser.add_argument('--cpu-concat', type=int, default=16)
    parser.add_argument('--n-samples', type=int, required=True)
    parser.add_argument('--buffer-region', type=int, default=250)
    parser.add_argument('--stages', type=str, default='impute,concat')
    parser.add_argument('--out-type', type=str, default='bcf', choices=['bcf', 'vcf'])
    parser.add_argument('--out-dir', required=True)

    args = parser.parse_args()

    genotype_imputation(input_vcf=args.input_vcf, females_file=args.samples_file, n_samples=args.n_samples,
                        buffer_region=args.buffer_region, phasing_software=args.phasing_software, local=args.local,
                        billing_project=args.billing_project, memory=args.memory, cpu=args.cpu_concat,
                        stages=args.stages, output_type=args.out_type, out_dir=args.out_dir)


if __name__ == '__main__':
    main()



# this is the script used to filter the reference 1KG+HGDP data used in PCA

import hail as hl

hl.init(default_reference='GRCh38')

ref_mt = hl.read_matrix_table('gs://african-seq-data/hgdp_tgp/hgdp_tgp_postQC.mt')

print("\nInitial number of SNPs before filtering: {}".format(ref_mt.count_rows()))
filtered_ref = hl.variant_qc(ref_mt)

print("filtering")

filtered_ref = filtered_ref.filter_rows((filtered_ref.variant_qc.AF[0] > 0.05) & (filtered_ref.variant_qc.AF[0] < 0.95))
print("\nNumber of SNPs after MAF filtering: {}".format(filtered_ref.count_rows()))

filtered_ref = filtered_ref.filter_rows(filtered_ref.variant_qc.call_rate > 0.999)
print("\nNumber of SNPs after Call Rate filtering: {}".format(filtered_ref.count_rows()))

# print("repartitioning")
# filtered_ref = filtered_ref.repartition(n_partitions=100, shuffle=True)

print("writing filtererd mt")
filtered_ref.write('gs://african-seq-data/hgdp_tgp/gwaspy_pca_ref/hgdp_1kg_filtered_maf_5_GRCh38.mt', overwrite=True)
print("Done filtering")

print("getting sample information")

mt = hl.read_matrix_table('gs://african-seq-data/hgdp_tgp/gwaspy_pca_ref/hgdp_1kg_filtered_maf_5_GRCh38.mt')

cols_ht = mt.cols()

pops = cols_ht.select(cols_ht.hgdp_tgp_meta.Study.region)

df = pops.to_pandas()

df.columns = ['Sample', 'SuperPop']

old_pops_labs = ['Africa', 'America', 'Central_South_Asia', 'East_Asia', 'Europe', 'Middle_East', 'Oceania', 'SAS']
new_pops_labs = ['AFR', 'AMR', 'CSA', 'EAS', 'EUR', 'MID', 'OCE', 'CSA']
df['SuperPop'] = df['SuperPop'].replace(old_pops_labs, new_pops_labs)

print(df['SuperPop'].value_counts())

print("exporting sample metadata")
df.to_csv('gs://african-seq-data/hgdp_tgp/gwaspy_pca_ref/hgdp_1kg_sample_info.tsv', sep='\t', index=False)




#!/usr/bin/env python3

__author__ = 'Lindo Nkambule'

import hail as hl


def bytes_to_gb(in_file: str):
    """
    Convert the size from bytes to GiB
    :param in_file: path to file, str
    :return: file size in GiB
    """

    file_info = hl.utils.hadoop_stat(in_file)
    size_bytes = file_info['size_bytes']
    size_gigs = size_bytes / (1024 * 1024 * 1024)

    return size_gigs



import hail as hl
from gwaspy.utils.sample_annotations import add_sample_annotations


def read_plink(dirname: str, basename: str) -> hl.MatrixTable:

    in_mt: hl.MatrixTable = hl.import_plink(bed=dirname + basename + '.bed',
                                            bim=dirname + basename + '.bim',
                                            fam=dirname + basename + '.fam',
                                            block_size=16)

    return in_mt


def read_vcf(dirname: str, basename: str) -> hl.MatrixTable:
    hl._set_flags(no_whole_stage_codegen='1')
    vcf_file = '{}{}.vcf.gz'.format(dirname, basename)
    hl.import_vcf(vcf_file, force_bgz=True, block_size=16).write('{}GWASpy.preimpQC.mt'.format(dirname), overwrite=True)
    in_mt = hl.read_matrix_table('{}GWASpy.preimpQC.mt'.format(dirname))

    # Unlike array data, a VCF might have multi-allelic sites
    # split multi-allelic sites into bi-allelic
    print("Checking for multi-allelic sites")
    pre_filt_multi_n = in_mt.count_rows()
    bi = in_mt.filter_rows(hl.len(in_mt.alleles) == 2)
    bi = bi.annotate_rows(a_index=hl.missing(hl.tint))  # when we update Hail version, use hl.missing instead of hl.null
    bi = bi.annotate_rows(was_split=False)

    multi = in_mt.filter_rows(hl.len(in_mt.alleles) > 2)
    split = hl.split_multi_hts(multi)

    in_mt = split.union_rows(bi)
    pos_filt_multi_n = in_mt.count_rows()
    print("Number of multi-allelic SNPs in VCF file: {}".format(pos_filt_multi_n-pre_filt_multi_n))

    return in_mt


def read_mt(dirname: str, basename: str) -> hl.MatrixTable:
    print(dirname + basename + ".mt")
    in_mt: hl.MatrixTable = hl.read_matrix_table(dirname + basename + ".mt")

    return in_mt


def read_infile(
        input_type: str = None,
        dirname: str = None, basename: str = None,
        **kwargs):

    global mt

    # vcf = kwargs.get('vcf')
    annotations = kwargs.get('annotations')

    if input_type == 'plink':
        mt = read_plink(dirname, basename)

    elif input_type == 'vcf':
        mt = read_vcf(dirname, basename)

    else:
        mt = read_mt(dirname, basename)

    if annotations:
        mt = add_sample_annotations(mt, annotations)

    return mt



__author__ = 'Lindo Nkambule'

import hail as hl
import pandas as pd
from gwaspy.pca.pca_filter_snps import pca_filter_mt, relatedness_check
import plotly.express as px


def pc_project(
        mt: hl.MatrixTable = None,
        loadings_ht: hl.Table = None,
        loading_location: str = 'loadings',
        af_location: str = 'pca_af') -> hl.Table:
    """
    Projects samples in `mt` on pre-computed PCs.

    :param mt: MT containing the samples to project
    :param loadings_ht: HT containing the PCA loadings and allele frequencies used for the PCA
    :param loading_location: Location of expression for loadings in `loadings_ht`
    :param af_location: Location of expression for allele frequency in `loadings_ht`
    :return: Table with scores calculated from loadings in column `scores`
    """

    n_variants = loadings_ht.count()

    mt = mt.annotate_rows(
        pca_loadings=loadings_ht[mt.row_key][loading_location],
        pca_af=loadings_ht[mt.row_key][af_location],
    )

    mt = mt.filter_rows(
        hl.is_defined(mt.pca_loadings)
        & hl.is_defined(mt.pca_af)
        & (mt.pca_af > 0)
        & (mt.pca_af < 1)
    )

    gt_norm = (mt.GT.n_alt_alleles() - 2 * mt.pca_af) / hl.sqrt(
        n_variants * 2 * mt.pca_af * (1 - mt.pca_af)
    )

    mt = mt.annotate_cols(scores=hl.agg.array_sum(mt.pca_loadings * gt_norm))

    return mt.cols().select('scores')


def intersect_ref(
        ref_dirname: str = 'gs://hgdp-1kg/hgdp_tgp/datasets_for_others/lindo/ds_without_outliers/',
        ref_basename: str = 'unrelated',
        data_mt: hl.MatrixTable = None,
        data_basename: str = None, out_dir: str = None):
    """
    Intersects reference panel with the data and writes intersections as matrix tables
    :param ref_dirname: directory name where reference data is
    :param ref_basename: base filename for reference data
    :param data_mt: input data MatrixTable
    :param data_basename: base filename for input data
    :param out_dir: output directory where files are going to be saved to
    :return:
    """
    print('Reading reference data mt')
    ref_mt = hl.read_matrix_table(f'{ref_dirname}{ref_basename}.mt')

    # filter data to sites in ref & array data
    data_in_ref = data_mt.filter_rows(hl.is_defined(ref_mt.rows()[data_mt.row_key]))
    print('\nsites in ref and data, inds in data: {}'.format(data_in_ref.count()))
    data_in_ref.write(f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/{data_basename}_intersect_1kg_hgdp.mt', overwrite=True)

    # filter ref to data sites
    ref_in_data = ref_mt.filter_rows(hl.is_defined(data_mt.rows()[ref_mt.row_key]))
    print('\nsites in ref and data, inds in ref: {}'.format(ref_in_data.count()))  #
    ref_in_data.write(f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/1kg_hgdp_intersect_{data_basename}.mt', overwrite=True)


def run_ref_pca(
        mt: hl.MatrixTable = None,
        npcs: int = 20,
        data_basename: str = None,
        out_dir: str = None):
    """
    Run PCA on a dataset
    :param mt: dataset to run PCA on
    :param npcs: number of principal components to be used in PCA
    :param data_basename: input data basename so outputs can be saved in correct dir
    :param out_dir: directory and filename prefix for where to put PCA output
    :return:
    """
    pca_evals, pca_scores, pca_loadings = hl.hwe_normalized_pca(mt.GT, k=npcs, compute_loadings=True)
    pca_mt = mt.annotate_rows(pca_af=hl.agg.mean(mt.GT.n_alt_alleles()) / 2)
    pca_loadings = pca_loadings.annotate(pca_af=pca_mt.rows()[pca_loadings.key].pca_af)

    # pca_scores.write(out_dir + 'GWASpy/PCA/' + '1000G_scores.ht', overwrite=True)
    # pca_scores = hl.read_table(out_dir + 'GWASpy/PCA/' + '1000G_scores.ht')
    pca_scores = pca_scores.transmute(**{f'PC{i}': pca_scores.scores[i - 1] for i in range(1, npcs+1)})
    pca_scores.export(f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/1kg_hgdp.project.pca.scores.txt.bgz')  # individual-level PCs

    pca_loadings.write(f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/1kg_hgdp_loadings.ht', overwrite=True)  # PCA loadings


def merge_data_with_ref(
        ref_scores: str = None,
        ref_info: str = None,
        data_scores: str = None) -> pd.DataFrame:
    """
    Merge data with ref
    :param ref_scores: path to reference score
    :param ref_info: path to information about samples in the ref scores
    :param data_scores: path to input data scores
    :return: a pandas Dataframe of data merged with reference
    """

    print('\nMerging data with ref')
    ref = pd.read_table(ref_scores, header=0, sep='\t', compression='gzip')
    data = pd.read_table(data_scores, header=0, sep='\t')
    ref_info = pd.read_table(ref_info, header=0, sep='\t')

    ref_merge = pd.merge(left=ref, right=ref_info, left_on='s', right_on='Sample', how='inner')

    data_ref = pd.concat([ref_merge, data], sort=False)
    print('\nDone merging data with ref')

    return data_ref


def plot_pca_ref(data_scores, ref_scores, ref_info, x_pc, y_pc):
    pcs = pd.read_table(data_scores, header=0, sep='\t')
    pcs['Project'] = "Input Dataset"
    pcs = pcs[['s', 'pop', 'Project', x_pc, y_pc]]

    ref = pd.read_table(ref_scores, header=0, sep='\t', compression='gzip')
    ref_info = pd.read_table(ref_info, header=0, sep='\t')

    ref_info.rename(columns={'Sample': 's'}, inplace=True)
    ref_update = pd.merge(ref, ref_info, how='left', on=['s'])
    ref_update.rename(columns={'SuperPop': 'pop'}, inplace=True)
    ref_update = ref_update[['s', 'pop', 'Project', x_pc, y_pc]]

    # concatenate the two dfs together
    concatenated = pd.concat([ref_update, pcs], axis=0)

    fig = px.scatter(concatenated, x=x_pc, y=y_pc, color='pop',
                     hover_data=['s', x_pc, y_pc, 'pop', 'Project'],
                     color_discrete_map={'AFR': "#984EA3", 'EAS': "#4DAF4A", 'EUR': "#377EB8", 'CSA': "#FF7F00",
                                         'AMR': "#E41A1C", 'MID': "#A65628", 'OCE': "#999999",
                                         'oth': "#F0E442"}).update_traces(marker_size=4)

    return fig


def run_pca_project(
        ref_dirname: str = 'gs://hgdp-1kg/hgdp_tgp/datasets_for_others/lindo/ds_without_outliers/',
        ref_basename: str = 'unrelated',
        ref_info: str = 'gs://hgdp-1kg/hgdp_tgp/gwaspy_pca_ref/hgdp_1kg_sample_info.unrelateds.pca_outliers_removed.with_project.tsv',
        data_dirname: str = None,
        data_basename: str = None,
        out_dir: str = None,
        input_type: str = None,
        reference: str = 'GRCh38',
        npcs: int = 20,
        maf: float = 0.05,
        hwe: float = 1e-3,
        call_rate: float = 0.98,
        ld_cor: float = 0.2,
        ld_window: int = 250000,
        relatedness_method: str = 'pc_relate',
        relatedness_thresh: float = 0.98,
        prob_threshold: float = 0.8):
    """
    Project samples into predefined PCA space
    :param ref_dirname: directory name where reference data is
    :param ref_basename: base filename for reference data
    :param ref_info: reference sample information
    :param data_dirname: matrix table of data to project
    :param data_basename: matrix table of data to project
    :param out_dir: directory and filename prefix for where to put PCA projection output
    :param input_type: input file(s) type: hail, plink, or vcf
    :param reference: reference build
    :param npcs: number of principal components to be used in PCA
    :param maf: minor allele frequency threshold
    :param hwe: hardy-weinberg fiter threshold
    :param call_rate: variant call rate filter threshold
    :param ld_cor: reference build
    :param ld_window: window size
    :param relatedness_method: method to use for relatedness filtering
    :param relatedness_thresh: threshold to use for filtering out related individuals
    :param prob_threshold: a list of probability thresholds to use for classifying samples
    :return: a pandas Dataframe with data PCA scores projected on the same PCA space using the Human Genome Diversity
    """
    print('\nReading data mt')
    if reference.lower() == 'grch37':
        lifted_over = f'{data_dirname}{data_basename}.liftover.grch38.mt'
        if not hl.hadoop_exists(lifted_over):
            from gwaspy.utils.reference_liftover import liftover_to_grch38
            mt = liftover_to_grch38(dirname=data_dirname, basename=data_basename, input_type=input_type)
        else:
            print(f'\nFound lifted-over over file: {lifted_over}')
            mt = hl.read_matrix_table(lifted_over)
    else:
        from gwaspy.utils.read_file import read_infile
        mt = read_infile(input_type=input_type, dirname=data_dirname, basename=data_basename)

    print('\nFiltering data mt')
    mt = pca_filter_mt(in_mt=mt, maf=maf, hwe=hwe, call_rate=call_rate, ld_cor=ld_cor, ld_window=ld_window)

    mt = relatedness_check(in_mt=mt, method=relatedness_method, outdir=out_dir, kin_estimate=relatedness_thresh)

    # Intersect data with reference
    intersect_ref(ref_dirname=ref_dirname, ref_basename=ref_basename, data_mt=mt, data_basename=data_basename,
                  out_dir=out_dir)

    ref_in_data = hl.read_matrix_table(f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/1kg_hgdp_intersect_{data_basename}.mt')

    print('\nComputing reference PCs')
    run_ref_pca(mt=ref_in_data, npcs=npcs, out_dir=out_dir, data_basename=data_basename)

    # project data
    pca_loadings = hl.read_table(f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/1kg_hgdp_loadings.ht')
    project_mt = hl.read_matrix_table(f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/{data_basename}_intersect_1kg_hgdp.mt')

    ht_projections = pc_project(mt=project_mt, loadings_ht=pca_loadings)
    ht_projections = ht_projections.transmute(**{f'PC{i}': ht_projections.scores[i - 1] for i in range(1, npcs+1)})
    ht_projections.export(f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/{data_basename}.project.pca.scores.tsv')

    ref_scores = f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/1kg_hgdp.project.pca.scores.txt.bgz'
    data_scores = f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/{data_basename}.project.pca.scores.tsv'
    data_ref = merge_data_with_ref(ref_scores=ref_scores, ref_info=ref_info, data_scores=data_scores)

    from gwaspy.pca.assign_pop_labels import assign_population_pcs
    pcs_df, clf = assign_population_pcs(pop_pc_pd=data_ref, num_pcs=npcs, min_prob=prob_threshold)

    data_pops = pcs_df.loc[pcs_df['SuperPop'].isnull()]
    data_pops['pop'].value_counts()
    cols = ['s', 'pop'] + [f'prob_{i}' for i in ["AFR", "AMR", "CSA", "EAS", "EUR", "MID", "OCE"]] + [f'PC{i}' for i in
                                                                                                      range(1, npcs+1)]
    data_pops_df = data_pops[cols]

    data_pops_df.to_csv(f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/pca_sup_pops_{prob_threshold}_probs.project.pca.txt',
                        sep='\t', index=False)

    print("\nGenerating PCA plots")
    data_scores_prob = f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/pca_sup_pops_{prob_threshold}_probs.project.pca.txt'

    figs_dict = {}
    # plotting more than 10 PCA plots in HTML generates wobbly, large files
    for i in range(1, 10, 2):
        xpc = f'PC{i}'
        ypc = f'PC{i + 1}'
        figs_dict["fig{}{}".format(xpc, ypc)] = plot_pca_ref(data_scores=data_scores_prob,
                                                             ref_scores=ref_scores,
                                                             ref_info=ref_info,
                                                             x_pc=xpc, y_pc=ypc)
    with open('/tmp/pca.project.plots.html', 'a') as f:
        for figname, figure in figs_dict.items():
            f.write(figure.to_html(include_plotlyjs='cdn'))

    hl.hadoop_copy('file:///tmp/pca.project.plots.html',
                   f'{out_dir}GWASpy/PCA/{data_basename}/pca_project/{data_basename}.pca.project.plots.html')






__author__ = 'Lindo Nkambule'

import hailtop.batch as hb
import hail as hl
import pandas as pd
from typing import Union
from gwaspy.utils.get_file_size import bytes_to_gb
from gwaspy.phasing.get_filebase import get_vcf_filebase


def aut_impute(b: hb.batch.Batch,
               vcf: hb.ResourceGroup = None,
               vcf_filename_no_ext: str = None,
               ref: hb.ResourceGroup = None,
               region: str = None,
               chromosome: str = None,
               buffer: int = 250,
               storage: int = None,
               memory: str = None,
               cpu: int = None,
               img: str = 'docker.io/lindonkambule/gwaspy:v1',
               out_dir: str = None):

    out_file_name = vcf_filename_no_ext + '.imputed.bcf'
    file_dir = vcf_filename_no_ext.split('.')[0]

    map_file = f'/shapeit4/maps/b38/{chromosome}.b38.gmap.gz'

    threads = cpu - 1

    impute = b.new_job(name=out_file_name)
    impute.cpu(cpu)
    impute.memory(memory)
    impute.storage(f'{storage}Gi')
    impute.image(img)

    cmd = f'''
        impute5_1.1.5_static \
            --h {ref.bcf} \
            --m {map_file} \
            --g {vcf.bcf} \
            --r {region} \
            --out-gp-field \
            --o {out_file_name} \
            --b {buffer} \
            --threads {threads}
    '''

    impute.command(cmd)
    # index file to use when merging
    impute.command(f'bcftools index {out_file_name}')

    impute.command(f'mv {out_file_name} {impute.ofile}')
    impute.command(f'mv {out_file_name}.csi {impute.idx}')
    b.write_output(impute.ofile, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}')
    b.write_output(impute.idx, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}.csi')


def sex_impute(b: hb.batch.Batch,
               vcf: hb.ResourceGroup = None,
               vcf_filename_no_ext: str = None,
               females_list: hb.ResourceFile = None,
               ref: hb.ResourceGroup = None,
               region: str = None,
               buffer: int = 250,
               storage: int = None,
               memory: str = None,
               cpu: int = None,
               img: str = 'docker.io/lindonkambule/gwaspy:v1',
               out_dir: str = None):

    global cmd

    in_females = females_list

    out_file_name = vcf_filename_no_ext + '.imputed.bcf'
    file_dir = vcf_filename_no_ext.split('.')[0]

    threads = cpu - 1

    impute = b.new_job(name=out_file_name)
    impute.cpu(cpu)
    impute.memory(memory)
    impute.storage(f'{storage}Gi')
    impute.image(img)

    start = int(region.split(':')[1].split('-')[0])
    end = int(region.split(':')[1].split('-')[1])

    # A. PAR1 REGION ONLY
    if end <= 2781479:
        # run diploid imputation
        impute.command('echo THIS CHUNK IS IN PAR1 REGION, SO WE WILL RUN DIPLOID IMPUTATION')
        map_file = '/shapeit4/maps/b38/chrX_par1.b38.gmap.gz'

        if start < 10001:
            start_new = 10001
            end_new = end
        else:
            start_new = start
            end_new = end

        new_imp_region = f'chrX:{start_new}-{end_new}'

        cmd = f'''
            impute5_1.1.5_static \
                --h {ref.bcf} \
                --m {map_file} \
                --g {vcf.bcf} \
                --r {new_imp_region} \
                --out-gp-field \
                --o {out_file_name} \
                --b {buffer} \
                --threads {threads}
        '''

        impute.command(cmd)
        # index file to use when merging
        impute.command(f'bcftools index {out_file_name}')

        impute.command(f'mv {out_file_name} {impute.ofile}')
        impute.command(f'mv {out_file_name}.csi {impute.idx}')
        b.write_output(impute.ofile, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}')
        b.write_output(impute.idx, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}.csi')

    # B. PAR2 REGION ONLY
    elif start >= 155701383:
        # run diploid imputation
        impute.command('echo THIS CHUNK IS IN PAR2 REGION, SO WE WILL RUN DIPLOID IMPUTATION')
        map_file = '/shapeit4/maps/b38/chrX_par2.b38.gmap.gz'

        if end > 156030895:
            end_new = 156030895
            start_new = start
        else:
            start_new = start
            end_new = end

        new_imp_region = f'chrX:{start_new}-{end_new}'

        cmd = f'''
            impute5_1.1.5_static \
                --h {ref.bcf} \
                --m {map_file} \
                --g {vcf.bcf} \
                --r {new_imp_region} \
                --out-gp-field \
                --o {out_file_name} \
                --b {buffer} \
                --threads {threads}
        '''

        impute.command(cmd)
        # index file to use when merging
        impute.command(f'bcftools index {out_file_name}')

        impute.command(f'mv {out_file_name} {impute.ofile}')
        impute.command(f'mv {out_file_name}.csi {impute.idx}')
        b.write_output(impute.ofile, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}')
        b.write_output(impute.idx, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}.csi')

    # C. NON-PAR REGION ONLY
    elif (start >= 2781479) & (end <= 155701382):
        map_file = '/shapeit4/maps/b38/chrX.b38.gmap.gz'
        start_new = start
        end_new = end
        new_imp_region = f'chrX:{start_new}-{end_new}'

        # (1) split by sex NB: THE USER SHOULD SUPPLY A FILE CONTAINING EITHER ONLY FEMALES OR MALE SAMPLE IDS
        # WITH ONE SAMPLE ID PER LINE
        cmd_split = f'''
                echo THIS CHUNK IS ONLY IN THE NON-PAR REGION, SO WE WILL SPLIT IT BY SEX BEFORE RUNNING IMPUTATION
                echo GETTING THE ORDER OF SAMPLES
                bcftools query -l {vcf.bcf} > samples_order.txt
                echo SPLITTING SAMPLES BY SEX
                bcftools view -S {in_females} {vcf.bcf} --output-type b --output females.bcf
                bcftools view -S ^{in_females} {vcf.bcf} --output-type b --output males.bcf
                echo INDEXING THE FILES BEFORE RUNNING IMPUTATION
                bcftools index females.bcf
                bcftools index males.bcf
        '''
        impute.command(cmd_split)

        # (2) run imputation separately for females and males
        cmd_impute = f'''
                impute5_1.1.5_static --h {ref.bcf} --m {map_file} --g females.bcf --r {new_imp_region} --out-gp-field \
                    --o females.imputed.bcf --b {buffer} --threads {threads}
                impute5_1.1.5_static --h {ref.bcf} --m {map_file} --g males.bcf --r {new_imp_region} --out-gp-field \
                    --o males.imputed.bcf --b {buffer} --threads {threads} -- haploid
        '''
        impute.command(cmd_impute)

        # (3) merge the imputed files back together into one chunk
        # (3a) index females and males files to be merged and create a file with these for bcftools
        cmd_index_chunks = f'''
                bcftools index females.imputed.bcf
                bcftools index males.imputed.bcf
                rm females.bcf* males.bcf*
        '''
        impute.command(cmd_index_chunks)

        # (3b) sometimes there are duplicates (raw and imputed with flipped alleles) and this causes an error when
        # merging, so we remove any duplicates
        cmd_sort = f'''
                echo CHECKING FOR DUPLICATE VARIANTS AND REMOVING THEM
                bcftools norm -d any females.imputed.bcf --output-type b --output females.imputed.sorted.bcf
                rm females.imputed.bcf*
                bcftools norm -d any males.imputed.bcf --output-type b --output males.imputed.sorted.bcf
                rm males.imputed.bcf*
                bcftools index females.imputed.sorted.bcf
                bcftools index males.imputed.sorted.bcf
                echo -e "females.imputed.sorted.bcf\nmales.imputed.sorted.bcf" > merge.txt
        '''

        impute.command(cmd_sort)

        # (3c) merge the files into one
        cmd_merge = f'''
                echo MERGING THE MALES AND FEMALES FILE BACK TOGETHER
                bcftools merge --file-list merge.txt --output-type b --output merged.sex.bcf
        '''
        impute.command(cmd_merge)

        # (3d) reorder samples back to how they were initially
        # splitting and re-merging samples will result in change in order compared to the initial file
        # so we have to use the initial samples order for the concat step by chromosome of imputation
        cmd_order = f'''
                echo ORDERING SAMPLES TO HOW THEY WERE INITIALLY
                bcftools view -S samples_order.txt merged.sex.bcf --output-type b --output {out_file_name}
        '''
        impute.command(cmd_order)

        impute.command(f'bcftools index {out_file_name}')

        impute.command(f'mv {out_file_name} {impute.ofile}')
        impute.command(f'mv {out_file_name}.csi {impute.idx}')
        b.write_output(impute.ofile, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}')
        b.write_output(impute.idx, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}.csi')

    # D. MIXED REGIONS
    else:
        if (start <= 2781479) & (end >= 2781479):
            map_file_par = '/shapeit4/maps/b38/chrX_par1.b38.gmap.gz'
            par_region = f'chrX:{start}-{2781479}'
            non_par_region = f'chrX:{2781479}-{end}'

            # to be used in step 3d
            cmd_merge_regs = f'''
                    echo ORDERING PAR1 SAMPLES TO HOW THEY WERE INITIALLY
                    bcftools view -S samples_order.txt par.imputed.bcf --output-type b --output par.imputed.sort.bcf
                    bcftools index par.imputed.sort.bcf
                    rm par.imputed.bcf*
                    echo MERGING THE PAR1 AND NON-PAR FILES BACK TOGETHER
                    echo -e "par.imputed.sort.bcf\nnonpar.imputed.sort.bcf" > merge_regions.txt
                    bcftools concat --naive --file-list merge_regions.txt --output-type b --output {out_file_name}
            '''

        else:
            map_file_par = '/shapeit4/maps/b38/chrX_par2.b38.gmap.gz'
            if end > 156030895:
                par_region = f'chrX:{155701383}-{156030895}'
            else:
                par_region = f'chrX:{155701383}-{end}'

            non_par_region = f'chrX:{start}-{155701382}'

            # to be used in step 3c
            cmd_merge_regs = f'''
                    echo ORDERING PAR2 SAMPLES TO HOW THEY WERE INITIALLY
                    bcftools view -S samples_order.txt par.imputed.bcf --output-type b --output par.imputed.sort.bcf
                    bcftools index par.imputed.sort.bcf
                    rm par.imputed.bcf*
                    echo MERGING THE PAR2 AND NON-PAR FILES BACK TOGETHER
                    echo -e "nonpar.imputed.sort.bcf\npar.imputed.sort.bcf" > merge_regions.txt
                    bcftools concat --naive --file-list merge_regions.txt --output-type b --output {out_file_name}
            '''

        cmd_split = f'''
                echo THIS CHUNK IS IN PAR1/2 and NON-PAR REGION, SO WE WILL SPLIT THESE TWO REGIONS
                echo GETTING THE ORDER OF SAMPLES
                bcftools query -l {vcf.bcf} > samples_order.txt
                echo SPLITTING OUT THE PAR REGION
                bcftools view {vcf.bcf} --regions {par_region} --output-type b --output par.bcf
                echo SPLITTING OUT THE NON-PAR REGION
                bcftools view {vcf.bcf} --regions {non_par_region} --output-type b --output nonpar.bcf
                echo SPLITTING THE NON-PAR REGION BY SEX
                bcftools view -S {in_females} nonpar.bcf --output-type b --output females.bcf
                bcftools view -S ^{in_females} nonpar.bcf --output-type b --output males.bcf
                echo INDEXING THE FILES BEFORE RUNNING IMPUTATION
                bcftools index par.bcf
                bcftools index females.bcf
                bcftools index males.bcf
                rm {vcf.bcf}
                rm nonpar.bcf
        '''

        impute.command(cmd_split)

        # (2) run imputation separately for par, females, and males
        map_file = '/shapeit4/maps/b38/chrX.b38.gmap.gz'
        cmd_impute = f'''
                impute5_1.1.5_static --h {ref.bcf} --m {map_file_par} --g par.bcf --r {par_region} --out-gp-field \
                    --o par.imputed.bcf --b {buffer} --threads {threads}
                impute5_1.1.5_static --h {ref.bcf} --m {map_file} --g females.bcf --r {non_par_region} --out-gp-field \
                    --o females.imputed.bcf --b {buffer} --threads {threads}
                impute5_1.1.5_static --h {ref.bcf} --m {map_file} --g males.bcf --r {non_par_region} --out-gp-field \
                    --o males.imputed.bcf --b {buffer} --threads {threads} -- haploid
        '''
        impute.command(cmd_impute)

        # (3) merge the imputed files back together into one chunk
        # (3a) index females and males files to be merged and create a file with these for bcftools
        cmd_index_chunks = f'''
                    bcftools index par.imputed.bcf
                    bcftools index females.imputed.bcf
                    bcftools index males.imputed.bcf
        '''
        impute.command(cmd_index_chunks)

        # (3b) sometimes there are duplicates (raw and imputed with flipped alleles) and this causes an error when
        # merging, so we remove any duplicates
        # bcftools view Nepal_PTSD_GSA_Updated_May2021_qced.chrX.phased.bcf | grep "#"
        cmd_sort = f'''
                echo CHECKING FOR DUPLICATE VARIANTS AND REMOVING THEM
                bcftools norm -d any females.imputed.bcf --output-type b --output females.imputed.sorted.bcf
                rm females.imputed.bcf*
                bcftools norm -d any males.imputed.bcf --output-type b --output males.imputed.sorted.bcf
                rm males.imputed.bcf*
                bcftools index females.imputed.sorted.bcf
                bcftools index males.imputed.sorted.bcf
                echo -e "females.imputed.sorted.bcf\nmales.imputed.sorted.bcf" > merge_sex.txt
        '''
        impute.command(cmd_sort)

        # (3c) merge the sex files into one
        cmd_merge_sex = f'''
                echo MERGING THE MALES AND FEMALES FILE BACK TOGETHER
                bcftools merge --file-list merge_sex.txt --output-type b --output nonpar.imputed.bcf
                bcftools index nonpar.imputed.bcf
                echo ORDERING NON-PAR SAMPLES TO HOW THEY WERE INITIALLY
                bcftools view -S samples_order.txt nonpar.imputed.bcf --output-type b --output nonpar.imputed.sort.bcf
                bcftools index nonpar.imputed.sort.bcf
                rm nonpar.imputed.bcf* females.imputed.sorted.bcf* males.imputed.sorted.bcf
        '''
        impute.command(cmd_merge_sex)

        # (3d) merge the non-par and par regions together
        impute.command(cmd_merge_regs)

        impute.command(f'bcftools index {out_file_name}')

        impute.command(f'mv {out_file_name} {impute.ofile}')
        impute.command(f'mv {out_file_name}.csi {impute.idx}')
        b.write_output(impute.ofile, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}')
        b.write_output(impute.idx, f'{out_dir}/GWASpy/{file_dir}/Imputation/imputed_chunks/{out_file_name}.csi')


def run_impute(backend: Union[hb.ServiceBackend, hb.LocalBackend] = None,
               input_vcf: str = None,
               females_file: str = None,
               n_samples: int = None,
               n_panel_samples: int = 4099,
               phasing_software: str = None,
               memory: str = 'highmem',
               buffer_region: int = 250,
               out_dir: str = None):

    global phased_bcf
    print(f'\n1. IMPUTATION ON {input_vcf} PHASED CHUNKS\n')
    vcf_filebase = get_vcf_filebase(input_vcf)

    impute_b = hb.Batch(backend=backend, name=f'impute-phased-chunks-{vcf_filebase}')

    # use regions file to update the regions for imputation so that there's no overlaps like in phasing
    regions = pd.read_csv(f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/refscatter.bed', delim_whitespace=True,
                          names=['chrom', 'start', 'end'])
    chroms_dfs = []

    for chrom, df_group in regions.groupby('chrom'):
        # print(df_group.loc[df_group.index[0], 'end'])
        df_group.loc[df_group.index[0], 'stop'] = df_group.loc[df_group.index[0], 'end']

        for i in range(1, len(df_group)):
            df_group.loc[df_group.index[i], 'stop'] = df_group.loc[df_group.index[i - 1], 'end'] + 1

        df_group['stop'] = df_group['stop'].astype(int)

        # add index column
        df_group['ind'] = df_group.index

        # update the first line to start at 1
        df_group.loc[df_group.index[0], 'stop'] = 1

        # combine the chromosome, start, and end positions into one
        df_group['reg'] = df_group['chrom'].astype(str) + ":" + df_group['stop'].astype(str) + "-" + df_group[
            'end'].astype(str)

        # select only the two needed columns
        regions_to_import_group = df_group[['reg', 'ind']]

        chroms_dfs.append(regions_to_import_group)

    regions_to_import = pd.concat(chroms_dfs, axis=0)
    regions_to_import = regions_to_import.sort_values('ind')
    regions_to_import.to_csv(f'{out_dir}/GWASpy/{vcf_filebase}/Imputation/imputation.regions', sep='\t', header=False,
                             index=False)

    regions_dict = pd.Series(regions_to_import.reg.values, index=regions_to_import.ind).to_dict()

    if phasing_software == 'shapeit':
        phased_vcfs_chunks = hl.utils.hadoop_ls(f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/phased_scatter/*.shapeit.bcf')
    else:
        phased_vcfs_chunks = hl.utils.hadoop_ls(f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/phased_scatter/*.eagle.bcf')

    for i in range(1, 24):
        if i == 23:
            chrom = 'chrX'
        else:
            chrom = f'chr{i}'

        ref_bcf = f'gs://gcp-public-data--gnomad/resources/hgdp_1kg/phased_haplotypes/hgdp.tgp.gwaspy.merged.{chrom}.merged.bcf'
        ref_size = bytes_to_gb(ref_bcf)
        ref = impute_b.read_input_group(**{'bcf': ref_bcf,
                                           'bcf.csi': f'{ref_bcf}.csi'})

        # output is not always bcf
        phased_filename = f'{out_dir}/GWASpy/{vcf_filebase}/Phasing/phased_merged/{vcf_filebase}.{chrom}.phased.{phasing_software}'
        if hl.hadoop_exists(f'{phased_filename}.bcf'):
            phased_bcf = f'{phased_filename}.bcf'
        elif hl.hadoop_exists(f'{phased_filename}.vcf.gz'):
            phased_bcf = f'{phased_filename}.vcf.gz'

        in_vcf = impute_b.read_input_group(**{'bcf': phased_bcf,
                                              'bcf.csi': f'{phased_bcf}.csi'})
        vcf_size = bytes_to_gb(input_vcf)

        disk_size = int(round(10.0 + 3.0 * vcf_size + ((1.0 + 2.0 * n_samples/n_panel_samples) * ref_size)))
        job_memory = memory
        job_cpu = 16 if job_memory == 'highmem' else 8

        for file in phased_vcfs_chunks:
            f = file['path']
            vcf_basename = get_vcf_filebase(f)
            file_index = int(vcf_basename.split('.')[-3])
            file_region = regions_dict[file_index]
            map_chrom = file_region.split(':')[0]

            imp_out_filename = f'{vcf_basename}.imputed.bcf'
            # file_dir = vcf_basename.split('.')[0]
            output_filepath_name = f'{out_dir}/GWASpy/{vcf_filebase}/Imputation/imputed_chunks/{imp_out_filename}'

            if map_chrom == chrom:
                # check if imputed file already exists
                if hl.hadoop_exists(output_filepath_name):
                    continue

                else:
                    if chrom == 'chrX':
                        females_in = impute_b.read_input(females_file)

                        sex_impute(b=impute_b, vcf=in_vcf, females_list=females_in, vcf_filename_no_ext=vcf_basename,
                                   ref=ref, region=file_region, buffer=buffer_region,
                                   storage=disk_size, memory=job_memory, cpu=job_cpu, out_dir=out_dir)
                    else:
                        aut_impute(b=impute_b, vcf=in_vcf, vcf_filename_no_ext=vcf_basename, ref=ref,
                                   region=file_region, chromosome=chrom, buffer=buffer_region, storage=disk_size,
                                   memory=job_memory, cpu=job_cpu, out_dir=out_dir)

    impute_b.run()




__author__ = 'Lindo Nkambule'

import hail as hl
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from gwaspy.pca.pca_filter_snps import pca_filter_mt, relatedness_check


def plot_pca(
        in_df: pd.DataFrame,
        x_pc: str,
        y_pc: str,
        column: str):

    fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(15, 15))

    if column == 'is_female':
        cat1 = in_df[in_df['is_female'] == 'female']
        label1 = 'female'
        color1 = 'blue'
        marker1 = "^"

        cat2 = in_df[in_df['is_female'] == 'male']
        label2 = 'male'
        color2 = 'orange'
        marker2 = "s"

        cat3 = in_df[in_df['is_female'] == 'unknown']
        label3 = 'unknown'
    else:
        cat1 = in_df[in_df['is_case'] == 'case']
        label1 = 'case'
        color1 = 'green'
        marker1 = "o"

        cat2 = in_df[in_df['is_case'] == 'control']
        label2 = 'control'
        color2 = 'darkorange'
        marker2 = "o"

        cat3 = in_df[in_df['is_case'] == 'unknown']
        label3 = 'unknown'

    axs.scatter(cat1[x_pc], cat1[y_pc], c=color1, label=label1, s=15, alpha=1, marker=marker1, edgecolor='black',
                linewidths=0.4)

    axs.scatter(cat2[x_pc], cat2[y_pc], c=color2, label=label2, s=15, alpha=1, marker=marker2, edgecolor='black',
                linewidths=0.4)

    axs.scatter(cat3[x_pc], cat3[y_pc], c='red', label=label3, s=15, alpha=1, marker='x', edgecolor='black',
                linewidths=0.4)

    plt.legend(bbox_to_anchor=(1, 0.5), loc='lower left', frameon=False, prop={'size': 13})
    axs.set_xlabel(xlabel=x_pc, fontsize=15)
    axs.set_ylabel(ylabel=y_pc, fontsize=15)

    plt.close()

    return fig


def run_pca_normal(
        dirname: str = None,
        basename: str = None,
        input_type: str = None,
        reference: str = 'GRCh38',
        maf: float = 0.05,
        hwe: float = 1e-3,
        call_rate: float = 0.98,
        ld_cor: float = 0.2,
        ld_window: int = 250000,
        n_pcs: int = 20,
        relatedness_method: str = 'pc_relate',
        relatedness_thresh: float = 0.98,
        out_dir: str = None):

    print('\nReading mt')
    if reference.lower() == 'grch37':
        lifted_over = f'{dirname}{basename}.liftover.grch38.mt'
        if not hl.hadoop_exists(lifted_over):
            from gwaspy.utils.reference_liftover import liftover_to_grch38
            mt = liftover_to_grch38(dirname=dirname, basename=basename, input_type=input_type)
        else:
            print(f'\nFound lifted-over over file: {lifted_over}')
            mt = hl.read_matrix_table(lifted_over)
    else:
        from gwaspy.utils.read_file import read_infile
        mt = read_infile(input_type=input_type, dirname=dirname, basename=basename)

    print('\nFiltering mt')
    mt = pca_filter_mt(in_mt=mt, maf=maf, hwe=hwe, call_rate=call_rate, ld_cor=ld_cor, ld_window=ld_window)

    mt = relatedness_check(in_mt=mt, method=relatedness_method, outdir=out_dir, kin_estimate=relatedness_thresh)

    pca_snps = mt.count_rows()
    if pca_snps > 1000000:
        import warnings
        warnings.warn(f'Too many SNPs to be used in PCA: {pca_snps}. This will make PCA run longer')

    print('\nRunning PCA')
    eigenvalues, pcs, _ = hl.hwe_normalized_pca(mt.GT, k=n_pcs)

    pcs_ht = pcs.transmute(**{f'PC{i}': pcs.scores[i - 1] for i in range(1, n_pcs+1)})

    # add phenotype and sex to the output, using information from the mt
    # first check if is_case and os_female fields exist in the mt
    all_column_field_names = list(mt.col)
    # sex status is a MUST but not phenotype status
    if 'is_case' in all_column_field_names:
        ann_cols = ['is_case', 'is_female']
    else:
        ann_cols = ['is_female']

    annotations_ht = mt.cols().select(*ann_cols)

    if 'is_case' in all_column_field_names:
        pcs_ht = pcs_ht.annotate(is_case=annotations_ht[pcs_ht.s].is_case)
    pcs_ht = pcs_ht.annotate(is_female=annotations_ht[pcs_ht.s].is_female)

    print('\nSaving PC scores file')
    out_scores_file = f'{out_dir}GWASpy/PCA/pca_normal/{basename}.pca.normal.scores.tsv'
    pcs_ht.export(out_scores_file)

    print('\nGenerating PCA plots')
    pcs_scores = pd.read_table(out_scores_file, header=0, sep='\t')

    if 'is_case' in all_column_field_names:
        pcs_scores[['is_case']] = pcs_scores[['is_case']].replace([True, False, None], ['case', 'control', 'unknown'])
    pcs_scores[['is_female']] = pcs_scores[['is_female']].replace([True, False, None], ['female', 'male', 'unknown'])

    figs_dict = {}
    for col in ann_cols:
        for i in range(1, n_pcs, 2):
            xpc = f'PC{i}'
            ypc = f'PC{i + 1}'

            figs_dict["fig{}{}".format(col, i)] = plot_pca(pcs_scores, xpc, ypc, col)

    pdf = PdfPages('/tmp/pca.no.ref.plots.pdf')
    for figname, figure in figs_dict.items():
        pdf.savefig(figure)
    pdf.close()
    hl.hadoop_copy('file:///tmp/pca.no.ref.plots.pdf',
                   f'{out_dir}GWASpy/PCA/pca_normal/{basename}.pca.no.ref.plots.pdf')

